"""Base class for objects of type connectivity."""
import logging

import numpy as np
from scipy.interpolate import interp2d

from vispy import scene
from vispy.scene import visuals
import vispy.visuals.transforms as vist

from .visbrain_obj import VisbrainObject
# from ..objects import ConnectObj
from ..io import download_file, is_sc_image_installed
from ..utils import (array2colormap, color2vb, mpl_cmap, normalize,
                     vpnormalize, vprecenter)

logger = logging.getLogger('visbrain')


class TopoObj(VisbrainObject):
    """Create a topoplot object.

    Parameters
    ----------
    name : string
        The name of the connectivity object.
    data : array_like
        Array of data of shape (n_channels)
    xyz : array_like | None
        Array of source's coordinates.
    channels : list | None
        List of channel names.
    system : {'cartesian', 'spherical'}
        Coordinate system.
    levels : array_like/int | None
        The levels at which the isocurve is constructed.
    level_colors : string/array_like | 'white'
        The color to use when drawing the line. If a list is given, it
        must be of shape (Nlev), if an array is given, it must be of
        shape (Nlev, ...). and provide one color per level
        (rgba, colorname). By default, all levels are whites.
    unit : {'degree', 'rad'}
        If system is 'spherical', specify if angles are in degrees or radians.
    line_color : array_like/string | 'black'
        Color of lines for the head, nose and eras.
    line_width : float | 4.
        Line width for the head, nose and eras.
    chan_size : float | 12.
        Size of channel names text.
    chan_mark_color : array_like/string | 'white'
        Color of channel markers.
    chan_mark_symbol : string | 'disc'
        Symbol to use for markers. Use disc, arrow, ring, clobber, square,
        diamond, vbar, hbar, cross, tailed_arrow, x, triangle_up,
        triangle_down, and star.
    chan_txt_color : array_like/string | 'black'
        Color of channel names.
    cmap : string | None
        Matplotlib colormap (like 'viridis', 'inferno'...).
    clim : tuple/list | None
        Colorbar limit. Every values under / over clim will
        clip.
    vmin : float | None
        Every values under vmin will have the color defined
        using the under parameter.
    vmax : float | None
        Every values over vmin will have the color defined
        using the over parameter.
    under : tuple/string | None
        Matplotlib color under vmin.
    over : tuple/string | None
        Matplotlib color over vmax.
    transform : VisPy.visuals.transforms | None
        VisPy transformation to set to the parent node.
    parent : VisPy.parent | None
        Line object parent.
    verbose : string
        Verbosity level.
    kw : dict | {}
        Optional arguments are used to control the colorbar
        (See :class:`ColorbarObj`).

    Notes
    -----
    List of supported shortcuts :

        * **s** : save the figure
        * **<delete>** : reset camera
    """

    ###########################################################################
    ###########################################################################
    #                                BUILT IN
    ###########################################################################
    ###########################################################################

    def __init__(self, name, data, xyz=None, channels=None, system='cartesian',
                 levels=None, level_colors='white', unit='degree',
                 line_color='black', line_width=3., chan_size=12.,
                 chan_offset=(0., 0., 0.), chan_mark_color='white',
                 chan_mark_symbol='disc', chan_txt_color='black',
                 cmap='viridis', clim=None, vmin=None, under='gray', vmax=None,
                 over='red', margin=.05, transform=None, parent=None,
                 verbose=None, **kw):
        """Init."""
        VisbrainObject.__init__(self, name, parent, transform, verbose, **kw)

        # ======================== VARIABLES ========================
        scale = 800.  # fix GL bugs for small plots
        pos = np.zeros((1, 3), dtype=np.float32)
        # Colors :
        line_color = color2vb(line_color)
        chan_txt_color = color2vb(chan_txt_color)
        self._chan_mark_color = color2vb(chan_mark_color)
        self._chan_mark_symbol = chan_mark_symbol
        # Disc interpolation :
        self._interp = .1
        self._pix = 64
        csize = int(self._pix / self._interp) if self._interp else self._pix
        l = csize / 2  # noqa

        # ======================== NODES ========================
        # Main topoplot node :
        self.node = scene.Node(name='Topoplot', parent=self._node)
        self.node.transform = vist.STTransform(scale=[scale] * 3)
        # Headset + channels :
        self.node_headfull = scene.Node(name='HeadChan', parent=self.node)
        # Headset node :
        self.node_head = scene.Node(name='Headset', parent=self.node_headfull)
        # Channel node :
        self.node_chan = scene.Node(name='Channels', parent=self.node_headfull)
        self.node_chan.transform = vist.STTransform(translate=(0., 0., -10.))
        # Dictionaries :
        kw_line = {'width': line_width, 'color': line_color,
                   'parent': self.node_head, 'antialias': False}

        # ======================== PARENT VISUALS ========================
        # Main disc :
        self.disc = visuals.Image(pos=pos, name='Disc', parent=self.node_head,
                                  interpolation='bilinear')

        # ======================== HEAD / NOSE / EAR ========================
        # ------------------ HEAD ------------------
        # Head visual :
        self.head = visuals.Line(pos=pos, name='Head', **kw_line)
        # Head circle :
        theta = np.arange(0, 2 * np.pi, 0.001)
        head = np.full((len(theta), 3), -1., dtype=np.float32)
        head[:, 0] = l * (1. + np.cos(theta))
        head[:, 1] = l * (1. + np.sin(theta))
        self.head.set_data(pos=head)

        # ------------------ NOSE ------------------
        # Nose visual :
        self.nose = visuals.Line(pos=pos, name='Nose', **kw_line)
        # Nose data :
        wn, hn = csize * 50. / 512., csize * 30. / 512.
        nose = np.array([[l - wn, 2 * l - wn, 2.],
                         [l, 2 * l + hn, 2.],
                         [l, 2 * l + hn, 2.],
                         [l + wn, 2 * l - wn, 2.]
                         ])
        self.nose.set_data(pos=nose, connect='segments')

        # ------------------ EAR ------------------
        we, he = csize * 10. / 512., csize * 30. / 512.
        ye = l + he * np.sin(theta)
        # Ear left data :
        self.earL = visuals.Line(pos=pos, name='EarLeft', **kw_line)
        # Ear left visual :
        ear_l = np.full((len(theta), 3), 3., dtype=np.float32)
        ear_l[:, 0] = 2 * l + we * np.cos(theta)
        ear_l[:, 1] = ye
        self.earL.set_data(pos=ear_l)

        # Ear right visual :
        self.earR = visuals.Line(pos=pos, name='EarRight', **kw_line)
        # Ear right data :
        ear_r = np.full((len(theta), 3), 3., dtype=np.float32)
        ear_r[:, 0] = 0. + we * np.cos(theta)
        ear_r[:, 1] = ye
        self.earR.set_data(pos=ear_r)

        # ================== CHANNELS ==================
        # Channel's markers :
        self.chan_markers = visuals.Markers(pos=pos, name='ChanMarkers',
                                            parent=self.node_chan)
        # Channel's text :
        self.chan_text = visuals.Text(pos=pos, name='ChanText',
                                      parent=self.node_chan, anchor_x='center',
                                      color=chan_txt_color,
                                      font_size=chan_size)

        # ================== CAMERA ==================
        self.rect = ((-scale / 2) * (1 + margin),
                     (-scale / 2) * (1 + margin),
                     scale * (1. + margin),
                     scale * (1.11 + margin))

        # ================== COORDINATES ==================
        auto = self._get_channel_coordinates(xyz, channels, system, unit)
        if auto:
            eucl = np.sqrt(self._xyz[:, 0]**2 + self._xyz[:, 1]**2).max()
            self.node_head.transform = vpnormalize(head, dist=2 * eucl)
            # Rescale between (-1:1, -1:1) = circle :
            circle = vist.STTransform(scale=(.5 / eucl, .5 / eucl, 1.))
            self.node_headfull.transform = circle
            # Text translation :
            tr = np.array([0., .8, 0.]) + np.array(chan_offset)
        else:
            # Get coordinates of references along the x and y-axis :
            ref_x, ref_y = self._get_ref_coordinates()
            # Recenter the topoplot :
            t = vist.ChainTransform()
            t.prepend(vprecenter(head))
            # Rescale (-ref_x:ref_x, -ref_y:ref_y) (ref_x != ref_y => ellipse)
            coef_x = 2 * ref_x / head[:, 0].max()
            coef_y = 2 * ref_y / head[:, 1].max()
            t.prepend(vist.STTransform(scale=(coef_x, coef_y, 1.)))
            self.node_head.transform = t
            # Rescale between (-1:1, -1:1) = circle :
            circle = vist.STTransform(scale=(.5 / ref_x, .5 / ref_y, 1.))
            self.node_headfull.transform = circle
            # Text translation :
            tr = np.array([0., .04, 0.]) + np.array(chan_offset)
        self.chan_text.transform = vist.STTransform(translate=tr)

        # ================== GRID INTERPOLATION ==================
        # Interpolation vectors :
        x = y = np.arange(0, self._pix, 1)
        xnew = ynew = np.arange(0, self._pix, self._interp)

        # Grid interpolation function :
        def _grid_interpolation(grid):
            f = interp2d(x, y, grid, kind='linear')
            return f(xnew, ynew)
        self._grid_interpolation = _grid_interpolation

        self.set_data(data, levels, level_colors, cmap, clim, vmin, under,
                      vmax, over)

    def __len__(self):
        """Return the number of channels."""
        return self._nchan

    def __bool__(self):
        """Return if coordinates exist."""
        return hasattr(self, '_xyz')

    def _get_camera(self):
        """Get the most adapted camera."""
        cam = scene.cameras.PanZoomCamera(rect=self.rect)
        cam.aspect = 1.
        return cam

    def set_data(self, data, levels=None, level_colors='white', cmap='viridis',
                 clim=None, vmin=None, under='gray', vmax=None, over='red'):
        """Set data to the topoplot.

        Parameters
        ----------
        data : array_like
            Array of data of shape (n_channels)
        levels : array_like/int | None
            The levels at which the isocurve is constructed.
        level_colors : string/array_like | 'white'
            The color to use when drawing the line. If a list is given, it
            must be of shape (Nlev), if an array is given, it must be of
            shape (Nlev, ...). and provide one color per level
            (rgba, colorname). By default, all levels are whites.
        cmap : string | None
            Matplotlib colormap (like 'viridis', 'inferno'...).
        clim : tuple/list | None
            Colorbar limit. Every values under / over clim will
            clip.
        vmin : float | None
            Every values under vmin will have the color defined
            using the under parameter.
        vmax : float | None
            Every values over vmin will have the color defined
            using the over parameter.
        under : tuple/string | None
            Matplotlib color under vmin.
        over : tuple/string | None
            Matplotlib color over vmax.
        """
        # ================== XYZ / CHANNELS / DATA ==================
        xyz = self._xyz[self._keeponly]
        channels = list(np.array(self._channels)[self._keeponly])
        data = np.asarray(data, dtype=float).ravel()
        if len(data) == len(self):
            data = data[self._keeponly]
        logger.info("    %i channels detected" % len(channels))

        # =================== CHANNELS ===================
        # Markers :
        radius = normalize(data, 10., 30.)
        self.chan_markers.set_data(pos=xyz, size=radius, edge_color='black',
                                   face_color=self._chan_mark_color,
                                   symbol=self._chan_mark_symbol)
        # Names :
        if channels is not None:
            self.chan_text.text = channels
            self.chan_text.pos = xyz

        # =================== GRID ===================
        pos_x, pos_y = xyz[:, 0], xyz[:, 1]
        xmin, xmax = pos_x.min(), pos_x.max()
        ymin, ymax = pos_y.min(), pos_y.max()
        xi = np.linspace(xmin, xmax, self._pix)
        yi = np.linspace(ymin, ymax, self._pix)
        xh, yi = np.meshgrid(xi, yi)
        grid = self._griddata(pos_x, pos_y, data, xh, yi)

        # =================== INTERPOLATION ===================
        if self._interp is not None:
            grid = self._grid_interpolation(grid)
        csize = max(self._pix, grid.shape[0])
        # Variables :
        l = csize / 2  # noqa
        y, x = np.ogrid[-l:l, -l:l]
        mask = x**2 + y**2 < l**2
        nmask = np.invert(mask)

        # =================== DISC ===================
        # Force min < off-disc values < max :
        d_min, d_max = data.min(), data.max()
        grid = normalize(grid, d_min, d_max)
        clim = (d_min, d_max) if clim is None else clim
        self._update_cbar_args(cmap, clim, vmin, vmax, under, over)
        grid_color = array2colormap(grid, **self.to_kwargs())
        grid_color[nmask, -1] = 0.
        # grid[nmask] = d_min
        # self.disc.clim = clim
        # self.disc.cmap = cmap_to_glsl(limits=(d_min, d_max),
        #                               translucent=(None, d_min),
        #                               **self.to_kwargs())
        self.disc.set_data(grid_color)

        # =================== LEVELS ===================
        if levels is not None:
            if isinstance(levels, int):
                levels = np.linspace(d_min, d_max, levels)
            if isinstance(level_colors, str):
                # Get colormaps :
                cmaps = mpl_cmap(bool(level_colors.find('_r') + 1))
                if level_colors in cmaps:
                    level_colors = array2colormap(levels, cmap=level_colors)
            grid[nmask] = np.inf
            is_sc_image_installed(True)
            self.iso = visuals.Isocurve(data=grid, parent=self.node_head,
                                        levels=levels, color_lev=level_colors,
                                        width=2.)
            self.iso.transform = vist.STTransform(translate=(0., 0., -5.))

    def connect(self, connect, **kwargs):
        """Draw connectivity lines between channels.

        Parameters
        ----------
        connect : array_like
            A 2D array of connectivity links of shape (n_channels, n_channels).
        kwargs : dict | {}
            Optional arguments are passed to the `visbrain.objects.ConnectObj`
            object.
        """
        logger.info("    Connect channels")
        # self._connect = ConnectObj('ChanConnect', self._xyz, connect,
        #                            parent=self.node_chan, **kwargs)

    def _get_channel_coordinates(self, xyz, channels, system, unit):
        """Get channel coordinates.

        Parameters
        ----------
        xyz : array_like | None
            Array of source's coordinates.
        channels : list | None
            List of channel names.
        system : {'cartesian', 'spherical'}
            Coordinate system.
        unit : string | {'degree', 'rad'}
            If system is 'spherical', specify if angles are in degrees or
            radians.
        """
        # =====================
        if (xyz is None) and (channels is None):  # Both None
            raise ValueError("You must either define sources using the xyz or"
                             " channels inputs")
        elif isinstance(xyz, np.ndarray):  # xyz exist
            if xyz.shape[1] not in [2, 3]:
                raise ValueError("Shape of xyz must be (nchan, 2) or "
                                 "(nchan, 3)")
            nchan = xyz.shape[0]
            if xyz.shape[1] == 2:
                xyz = np.c_[xyz, np.zeros((nchan), dtype=np.float)]
            xyz[:, 2] = 1.
            keeponly = np.ones((xyz.shape[0],), dtype=bool)
            channels = [''] * nchan if channels is None else channels
            auto = True
        elif (xyz is None) and (channels is not None):  # channels exist
            if all([isinstance(k, str) for k in channels]):
                xyz, keeponly = self._get_coordinates_from_name(channels)
                system, unit = 'spherical', 'degree'
                auto = False

        # Select channels to use :
        if any(keeponly):
            if not all(keeponly):
                ignore = list(np.array(channels)[np.invert(keeponly)])
                logger.warning("Ignored channels for topoplot :"
                               " %s" % ', '.join(ignore))

            # ----------- Conversion -----------
            if isinstance(xyz, np.ndarray):
                if system == 'cartesian':
                    pass  # all good
                elif system == 'spherical':
                    xyz = self._spherical_to_cartesian(xyz, unit)
                    xyz = self._array_project_radial_to3d(xyz)

        self._xyz = xyz
        self._channels = channels
        self._keeponly = keeponly
        self._nchan = len(channels)

        return auto

    def _get_ref_coordinates(self, x='T4', y='Fpz'):
        """Get cartesian coordinates for electrodes to use as references.

        The ELAN software use by default spherical coordinates with T4 as the
        extrema for the x-axis and Fpz as the extrema for the y-axis.

        Parameters
        ----------
        x : string | 'T4'
            Name of the electrode t use as a reference for the x-axis.
        y : string | 'Fpz'
            Name of the electrode t use as a reference for the y-axis.
        """
        ref = self._get_coordinates_from_name([x, y])[0]
        ref = self._spherical_to_cartesian(ref, unit='degree')
        ref = self._array_project_radial_to3d(ref)
        ref_x, ref_y = ref[0, 0], ref[1, 1]
        return ref_x, ref_y

    @staticmethod
    def _get_coordinates_from_name(chan):
        """From the name of the channels, find xyz coordinates.

        Parameters
        ----------
        chan : list
            List of channel names.
        """
        # Load the coordinates template :
        path = download_file('eegref.npz', astype='topo')
        file = np.load(path)
        name_ref, xyz_ref = file['chan'], file['xyz']
        keeponly = np.ones((len(chan)), dtype=bool)
        # Find and load xyz coordinates :
        xyz = np.zeros((len(chan), 3), dtype=np.float32)
        for num, k in enumerate(chan):
            # Find if the channel is present :
            idx = np.where(name_ref == k.lower())[0]
            if idx.size:
                xyz[num, 0:2] = np.array(xyz_ref[idx[0], :])
            else:
                keeponly[num] = False

        return np.array(xyz), keeponly

    @staticmethod
    def _spherical_to_cartesian(xyz, unit='rad'):
        """Convert spherical coordinates to cartesian.

        Parameters
        ----------
        xyz : array_like
            The array of spheric coordinate of shape (N, 3).
        unit : {'rad', 'degree'}
            Specify the unit angles.

        Returns
        -------
        xyz : array_like
            The cartesian coordinates of the angle of shape (N, 3).
        """
        # Get theta / phi :
        theta, phi = xyz[:, 0], xyz[:, 1]
        if unit is 'degree':
            np.deg2rad(theta, out=theta)
            np.deg2rad(phi, out=phi)
        # Get radius :
        r = np.sin(theta)
        # Get cartesian coordinates :
        np.multiply(np.cos(phi), r, out=xyz[:, 0])
        np.multiply(np.sin(phi), r, out=xyz[:, 1])
        np.cos(theta, xyz[:, 2])
        return xyz

    @staticmethod
    def _griddata(x, y, v, xi, yi):
        """Make griddata."""
        xy = x.ravel() + y.ravel() * -1j
        d = xy[None, :] * np.ones((len(xy), 1))
        d = np.abs(d - d.T)
        n = d.shape[0]
        d.flat[::n + 1] = 1.

        g = (d * d) * (np.log(d) - 1.)
        g.flat[::n + 1] = 0.
        weights = np.linalg.solve(g, v.ravel())

        m, n = xi.shape
        zi = np.zeros_like(xi)
        xy = xy.T

        g = np.empty(xy.shape)
        for i in range(m):
            for j in range(n):
                d = np.abs(xi[i, j] + -1j * yi[i, j] - xy)
                mask = np.where(d == 0)[0]
                if len(mask):
                    d[mask] = 1.
                np.log(d, out=g)
                g -= 1.
                g *= d * d
                if len(mask):
                    g[mask] = 0.
                zi[i, j] = g.dot(weights)
        return zi

    @staticmethod
    def _array_project_radial_to3d(points_2d):
        """Radial 3d projection."""
        points_2d = np.atleast_2d(points_2d)
        alphas = np.sqrt(np.sum(points_2d**2, -1))

        betas = np.sin(alphas) / alphas
        betas[alphas == 0] = 1
        x = points_2d[..., 0] * betas
        y = points_2d[..., 1] * betas
        z = np.cos(alphas)

        points_3d = np.asarray([x, y, z]).T

        return points_3d



"""Main class for settings managment."""
import numpy as np
import datetime
from PyQt5.QtCore import QObjectCleanupHandler

import vispy.visuals.transforms as vist


class UiSettings(object):
    """Main class for settings managment."""

    def __init__(self):
        """Init."""
        # =====================================================================
        # SLIDER
        # =====================================================================
        self._slFrame.setMaximumHeight(100)
        # Function applied when the slider move :
        self._slOnStart = False
        self._fcn_slider_settings()
        self._SlVal.valueChanged.connect(self._fcn_slider_move)
        # Function applied when slider's settings changed :
        self._SigWin.valueChanged.connect(self._fcn_slider_move)
        self._SigWin.setKeyboardTracking(False)
        self._SigSlStep.valueChanged.connect(self._fcn_slider_settings)
        self._SigSlStep.setKeyboardTracking(False)
        # Spin box for window selection :
        self._SlGoto.valueChanged.connect(self._fcn_slider_win_selection)
        self._SlGoto.setKeyboardTracking(False)
        # Unit conversion :
        self._slRules.currentIndexChanged.connect(self._fcn_slider_move)
        # Grid toggle :
        self._slGrid.clicked.connect(self._fcn_grid_toggle)
        # (port-visbrain) Grid scale change :
        self._SignV.valueChanged.connect(self._fcn_grid_scale)
        # (port-visbrain) Channel stage coloring toggle :
        self._slStageColors.clicked.connect(self._fcn_chan_stage_colors_toggle)
        # Text format :
        self._slTxtFormat = "Window : [ {start} ; {end} ] {unit} || " + \
                            "Sleep stage : {conv}"
        # Absolute time :
        self._slAbsTime.clicked.connect(self._fcn_slider_move)
        # Magnify :
        self._slMagnify.clicked.connect(self._fcn_slider_magnify)
        # Annotation from the navigation bar :
        self._AnnotateRun.clicked.connect(self._fcn_annotate_nav)

    # =====================================================================
    # SLIDER
    # =====================================================================
    def _fcn_slider_move(self):
        """Function applied when the slider move."""
        # ================= INDEX =================
        # Get slider variables :
        val = self._SlVal.value()
        step = self._SigSlStep.value()
        win = self._SigWin.value()
        xlim = (val * step, val * step + win)
        iszoom = self.menuDispZoom.isChecked()
        unit = str(self._slRules.currentText())
        # Find closest time index :
        t = [0, 0]
        t[0] = int(round(np.abs(self._time - xlim[0]).argmin()))
        t[1] = int(round(np.abs(self._time - xlim[1]).argmin()))
        # Hypnogram info :
        hypref = int(self._hypno[t[0]])
        hypconv = self._hconv[hypref]
        hypcol = self._hypcolor[hypconv]
        stage = str(self._hypYLabels[hypconv + 2].text())

        # ================= MESH UPDATES =================
        # ---------------------------------------
        # Update display signal :
        sl = slice(t[0], t[1])
        self._chan.set_data(self._sf, self._data, self._hypno, self._time, sl=sl,
                            ylim=self._ylims)

        # ---------------------------------------
        is_indic_checked = self.menuDispIndic.isChecked()
        # Update spectrogram indicator :
        if is_indic_checked and not iszoom:
            ylim = (self._PanSpecFstart.value(), self._PanSpecFend.value())
            self._specInd.set_data(xlim=xlim, ylim=ylim)

        # ---------------------------------------
        # Update hypnogram indicator :
        if is_indic_checked and not iszoom:
            self._hypInd.set_data(xlim=xlim, ylim=(-6., 2.))

        # ---------------------------------------
        # Update topoplot if visible :
        if self._topoW.isVisible():
            # Prepare data before plotting :
            data = self._topo._prepare_data(self._sf, self._data[:, sl],
                                            self._time[sl]).mean(1)
            # Set preprocessed sleep data :
            self._topo.set_sleep_topo(data)
            # Update title :
            fm, fh = self._PanTopoFmin.value(), self._PanTopoFmax.value()
            dispas = self._PanTopoDisp.currentText()
            txt = 'Mean ' + dispas + ' in\n[' + str(fm) + ';' + str(fh) + 'hz]'
            self._topoTitle.setText(txt)
            self._topoTitle.setStyleSheet("QLabel {color: " +
                                          hypcol + ";}")

        # ---------------------------------------
        # Update Time indicator :
        if is_indic_checked:
            self._TimeAxis.set_data(xlim[0], win, self._time, unit=unit,
                                    markers=self._annot_mark)

        # ================= GUI =================
        # Update Go to :
        self._SlGoto.setValue(val * step)

        # ================= ZOOMING =================
        if iszoom:
            xlim_diff = xlim[1] - xlim[0]
            # Histogram :
            self._hypcam.rect = (xlim[0], -5, xlim_diff, 7.)
            # Spectrogram :
            self._speccam.rect = (xlim[0], self._spec.freq[0], xlim_diff,
                                  self._spec.freq[-1] - self._spec.freq[0])
            # Time axis :
            self._TimeAxis.set_data(xlim[0], win, np.array([xlim[0], xlim[1]]),
                                    unit='seconds', markers=self._annot_mark)
            self._timecam.rect = (xlim[0], 0., win, 1.)

        # ================= TEXT INFO =================
        # Get unit and convert:
        if self._slAbsTime.isChecked():
            xlim = np.asarray(xlim) + self._toffset
            start = str(datetime.datetime.utcfromtimestamp(
                xlim[0])).split(' ')[1]
            stend = str(datetime.datetime.utcfromtimestamp(
                xlim[1])).split(' ')[1]
            txt = "Window : [ " + start + " ; " + stend + " ] || Sleep " + \
                "stage : " + stage
        else:
            unit = self._slRules.currentText()
            if unit == 'seconds':
                fact = 1.
            elif unit == 'minutes':
                fact = 60.
            elif unit == 'hours':
                fact = 3600.
            xconv = np.round(1000. * np.array(xlim) / fact) / 1000.
            # Format string :
            txt = self._slTxtFormat.format(start=str(xconv[0]),
                                           end=str(xconv[1]), unit=unit,
                                           conv=stage)
        # Set text :
        self._SlText.setText(txt)
        self._SlText.setFont(self._font)
        self._SlText.setStyleSheet("QLabel {color: " +
                                   hypcol + ";}")

        # ================= HYPNO LABELS =================
        for k in self._hypYLabels:
            k.setStyleSheet("QLabel")
        self._hypYLabels[hypconv + 2].setStyleSheet("QLabel {color: " +
                                                    hypcol + ";}")

    def _fcn_slider_settings(self):
        """Function applied to change slider settings."""
        # Get current slider value :
        sl = self._SlVal.value()
        slmax = self._SlVal.maximum()
        win = self._SigWin.value()
        # TODO: casted all inputs to int because python was complaining
        # Set minimum : 
        self._SlVal.setMinimum(int(self._time.min()))
        # Set maximum : 
        step = self._SigSlStep.value()
        self._SlVal.setMaximum(int(((self._time.max() - win) / step) + 1))
        self._SlVal.setTickInterval(int(step))
        self._SlVal.setSingleStep(int(step))
        self._SlGoto.setMaximum(int((self._time.max() - win)))
        # Re-set slider value :
        self._SlVal.setValue(int(sl * self._SlVal.maximum() / slmax))

        if self._slOnStart:
            self._fcn_slider_move()
            # Update grid :
            if self.menuDispZoom.isChecked():
                self._hyp.set_grid(self._time, step)
            else:
                self._hyp.set_grid(self._time, win)
        else:
            self._slOnStart = True

    def _fcn_slider_win_selection(self):
        """Move slider using window spin."""
        # TODO: casted to int
        self._SlVal.setValue(int(self._SlGoto.value() / self._SigSlStep.value()))

    def _fcn_slider_magnify(self):
        """Magnify signals."""
        # Set transformation to each node parent :
        for k in self._chan.node:
            # Use either Magnify / Null transformation :
            if self._slMagnify.isChecked():
                transform = vist.nonlinear.Magnify1DTransform()
            else:
                transform = vist.NullTransform()
            k.transform = transform

    # =====================================================================
    # GRID
    # =====================================================================
    def _fcn_grid_toggle(self):
        """Toggle grid visibility."""
        viz = self._slGrid.isChecked()
        # Toggle hypno grid :
        self._hyp.grid.visible = viz
        # Toggle grid for each channel :
        for k in self._chan.grid:
            k.visible = viz

    # (port-visbrain)
    def _fcn_grid_scale(self, value):
        scale = (1., value)
        for k in self._chan.grid:
            k.transform = vist.STTransform(scale=scale)
            # k.scale = scale

    # =====================================================================
    # RULER
    # =====================================================================
    def _get_fact_from_unit(self):
        """Get factor conversion from current selected unit."""
        unit = str(self._slRules.currentText())
        if unit == 'seconds':
            fact = 1.
        elif unit == 'minutes':
            fact = 60.
        elif unit == 'hours':
            fact = 3600.
        return fact

    def on_mouse_wheel(self, event):
        """Executed function on mouse wheel."""
        # TODO: casted to int
        self._SlVal.setValue(int(self._SlVal.value() + event.delta[1]))

    # =====================================================================
    # HYPNO
    # =====================================================================
    def _add_stage_on_win(self, stage):
        """Change the stage on the current window."""
        # Get the window :
        win = self._SigWin.value()
        val = self._SlVal.value()
        step = self._SigSlStep.value()
        xlim = (val * step, val * step + win)
        # Find closest time index :
        t = [0, 0]
        t[0] = int(round(np.abs(self._time - xlim[0]).argmin()))
        t[1] = int(round(np.abs(self._time - xlim[1]).argmin()))
        # Set the stage :
        self._hypno[t[0]:t[1]] = stage
        self._hyp.set_stage(t[0], t[1], stage)
        # # Update info table :
        self._fcn_info_update()
        # Update scoring table :
        self._fcn_hypno_to_score()
        # self._fcn_score_to_hypno()

    def _fcn_chan_stage_colors_toggle(self):
        viz = self._slStageColors.isChecked()
        self._chan.stage_colors_enabled = viz
        self._fcn_slider_move()

    # =====================================================================
    # Annotate
    # =====================================================================
    def _fcn_annotate_nav(self):
        """Annotate from the selected window."""
        # Set the current tab to the annotation tab :
        self.QuickSettings.setCurrentIndex(5)
        # Run annotation :
        self._fcn_annotate_add('')

    # =====================================================================
    # CLEAN / RESET GUI
    # =====================================================================
    def _fcn_clean_gui(self):
        """Clean the entire GUI."""
        # -------------- TABLES --------------
        # Info :
        self._infoTable.clear()
        self._infoTable.setRowCount(0)

        # Detection :
        self._DetectLocations.clear()
        self._DetectLocations.setRowCount(0)

        # -------------- LIST BOX --------------
        # Disconnect :
        self._PanSpecChan.currentIndexChanged.disconnect()
        # Clear items :
        self._PanSpecChan.clear()
        self._ToolDetectChan.clear()

        # -------------- VISUALS --------------
        # Channels :
        for k in range(len(self)):
            # Disconnect buttons :
            self._chanChecks[k].clicked.disconnect()
            self._yminSpin[k].valueChanged.disconnect()
            self._ymaxSpin[k].valueChanged.disconnect()
            # Delete elements :
            self._chanChecks[k].deleteLater()
            self._yminSpin[k].deleteLater(), self._ymaxSpin[k].deleteLater()
            self._chanWidget[k].deleteLater()
            self._chanLayout[k].deleteLater()
            self._chanLabels[k].deleteLater()
            self._amplitudeTxt[k].deleteLater()
            self._chanCanvas[k].parent = None
        QObjectCleanupHandler().add(self._chanGrid)
        QObjectCleanupHandler().clear()
        # Spectrogram :
        self._specCanvas.parent = None
        self._SpecW.deleteLater(), self._SpecLayout.deleteLater()
        self._specLabel.deleteLater()
        # Hypnogram :
        self._hypCanvas.parent = None
        self._HypW.deleteLater(), self._HypLayout.deleteLater()
        self._hypLabel.deleteLater()
        # Time axis :
        self._TimeAxis.parent = None
        self._TimeAxisW.deleteLater(), self._TimeLayout.deleteLater()
        self._timeLabel.deleteLater()

    def _fcn_reset_gui(self):
        """Reset the GUI."""
        from .uiElements import uiElements
        from ...visuals import visuals
        from ...tools import Tools
        uiElements.__init__(self)
        self._cam_creation()
        visuals.__init__(self)
        Tools.__init__(self)
        self._fcns_on_creation()



"""write GUI tables.

- write_table_txt : as text file
- write_table_csv : as csv file
"""

__all__ = ['write_table_txt', 'write_table_csv']


def write_table_txt():
    """Write a table in a text (txt) file."""
    pass


def write_table_csv():
    """Write a table in a CSV (csv) file."""
    pass



"""Group functions for color managment.

This file contains a bundle of functions that can be used to have a more
flexible control of diffrent problem involving colors (like turn an array /
string / faces into RBGA colors, defining the basic colormap object...)
"""
import logging

import numpy as np

from vispy.color.colormap import Colormap as VispyColormap

from matplotlib import cm
import matplotlib.colors as mplcol
from warnings import warn

from .sigproc import normalize
from .mesh import vispy_array


__all__ = ('Colormap', 'color2vb', 'array2colormap', 'cmap_to_glsl',
           'dynamic_color', 'color2faces', 'type_coloring', 'mpl_cmap',
           'color2tuple', 'mpl_cmap_index')


logger = logging.getLogger('visbrain')


class Colormap(object):
    """Main colormap class.

    Parameters
    ----------
    cmap : string | inferno
        Matplotlib colormap
    clim : tuple/list | None
        Limit of the colormap. The clim parameter must be a tuple / list
        of two float number each one describing respectively the (min, max)
        of the colormap. Every values under clim[0] or over clim[1] will
        peaked.
    alpha : float | 1.0
        The opacity to use. The alpha parameter must be between 0 and 1.
    vmin : float | None
        Threshold from which every color will have the color defined using
        the under parameter bellow.
    under : tuple/string | 'dimgray'
        Matplotlib color for values under vmin.
    vmax : float | None
        Threshold from which every color will have the color defined using
        the over parameter bellow.
    over : tuple/string | 'darkred'
        Matplotlib color for values over vmax.
    translucent : tuple | None
        Set a specific range translucent. With f_1 and f_2 two floats, if
        translucent is :

            * (f_1, f_2) : values between f_1 and f_2 are set to translucent
            * (None, f_2) x <= f_2 are set to translucent
            * (f_1, None) f_1 <= x are set to translucent
    lut_len : int | 1024
        Number of levels for the colormap.
    interpolation : {None, 'linear', 'cubic'}
        Interpolation type. Default is None.

    Attributes
    ----------
    data : array_like
        Color data of shape (n_data, 4)
    shape : tuple
        Shape of the data.
    r : array_like
        Red levels.
    g : array_like
        Green levels.
    b : array_like
        Blue levels.
    rgb : array_like
        RGB levels.
    alpha : array_like
        Transparency level.
    glsl : vispy.colors.Colormap
        GL colormap version.
    """

    def __init__(self, cmap='viridis', clim=None, vmin=None, under=None,
                 vmax=None, over=None, translucent=None, alpha=1.,
                 lut_len=1024, interpolation=None):
        """Init."""
        # Keep color parameters into a dict :
        self._kw = dict(cmap=cmap, clim=clim, vmin=vmin, vmax=vmax,
                        under=under, over=over, translucent=translucent,
                        alpha=alpha)
        # Color conversion :
        if isinstance(cmap, np.ndarray):
            assert (cmap.ndim == 2) and (cmap.shape[-1] in (3, 4))
            # cmap = single color :
            if (cmap.shape[0] == 1) and isinstance(interpolation, str):
                logger.debug("Colormap : unique color repeated.")
                data = np.tile(cmap, (lut_len, 1))
            elif (cmap.shape[0] == lut_len) or (interpolation is None):
                logger.debug("Colormap : Unique repeated.")
                data = cmap
            else:
                from scipy.interpolate import interp2d
                n_ = cmap.shape[1]
                x, y = np.linspace(0, 1, n_), np.linspace(0, 1, cmap.shape[0])
                f = interp2d(x, y, cmap, kind=interpolation)
                # Interpolate colormap :
                data = f(x, np.linspace(0, 1, lut_len))
        elif isinstance(cmap, str):
            data = array2colormap(np.linspace(0., 1., lut_len), **self._kw)
        # Alpha correction :
        if data.shape[-1] == 3:
            data = np.c_[data, np.full((data.shape[0],), alpha)]
        # NumPy float32 conversion :
        self._data = vispy_array(data)

    def to_rgba(self, data):
        """Turn a data vector into colors using colormap properties.

        Parameters
        ----------
        data : array_like
            Vector of data of shape (n_data,).

        Returns
        -------
        color : array_like
            Array of colors of shape (n_data, 4)
        """
        if isinstance(self._kw['cmap'], np.ndarray):
            return self._data
        else:
            return array2colormap(data, **self._kw)

    def __len__(self):
        """Get the number of colors in the colormap."""
        return self._data.shape[0]

    def __getitem__(self, name):
        """Get a color item."""
        return self._kw[name]

    @property
    def data(self):
        """Get colormap data."""
        return self._data

    @property
    def shape(self):
        """Get the shape of the data."""
        return self._data.shape

    @property
    def glsl(self):
        """Get a glsl version of the colormap."""
        return cmap_to_glsl(lut_len=len(self), **self._kw)

    @property
    def r(self):
        """Get red levels."""
        return self._data[:, 0]

    @property
    def g(self):
        """Get green levels."""
        return self._data[:, 1]

    @property
    def b(self):
        """Get blue levels."""
        return self._data[:, 2]

    @property
    def rgb(self):
        """Get rgb levels."""
        return self._data[:, 0:3]

    @property
    def alpha(self):
        """Get transparency level."""
        return self._data[:, -1]


def color2vb(color=None, default=(1., 1., 1.), length=1, alpha=1.0,
             faces_index=False):
    """Turn into a RGBA compatible color format.

    This function can tranform a tuple of RGB, a matplotlib color or an
    hexadecimal color into an array of RGBA colors.

    Parameters
    ----------
    color : None/tuple/string | None
        The color to use. Can either be None, or a tuple (R, G, B),
        a matplotlib color or an hexadecimal color '#...'.
    default : tuple | (1,1,1)
        The default color to use instead.
    length : int | 1
        The length of the output array.
    alpha : float | 1
        The opacity (Last digit of the RGBA tuple).
    faces_index : bool | False
        Specify if the returned color have to be compatible with faces index
        (e.g a (n_color, 3, 4) array).

    Return
    ------
    vcolor : array_like
        Array of RGBA colors of shape (length, 4).
    """
    # Default or static color :
    if (color is None) or isinstance(color, (str, tuple, list, np.ndarray)):
        if color is None:  # Default
            coltuple = default
        elif isinstance(color, (tuple, list, np.ndarray)):  # Static
            color = np.squeeze(color).ravel()
            if len(color) == 4:
                alpha = color[-1]
                color = color[0:-1]
            coltuple = color
        elif isinstance(color, str) and (color[0] != '#'):  # Matplotlib
            # Check if the name is in the Matplotlib database :
            if color in mplcol.cnames.keys():
                coltuple = mplcol.hex2color(mplcol.cnames[color])
            else:
                warn("The color name " + color + " is not in the matplotlib "
                     "database. Default color will be used instead.")
                coltuple = default
        elif isinstance(color, str) and (color[0] == '#'):  # Hexadecimal
            try:
                coltuple = mplcol.hex2color(color)
            except:
                warn("The hexadecimal color " + color + " is not valid. "
                     "Default color will be used instead.")
                coltuple = default
        # Set the color :
        vcolor = np.concatenate((np.array([list(coltuple)] * length),
                                 alpha * np.ones((length, 1),
                                                 dtype=np.float32)), axis=1)

        # Faces index :
        if faces_index:
            vcolor = np.tile(vcolor[:, np.newaxis, :], (1, 3, 1))

        return vcolor.astype(np.float32)
    else:
        raise ValueError(str(type(color)) + " is not a recognized type of "
                         "color. Use None, tuple or string")


def color2tuple(color, astype=np.float32, rmalpha=True, roundto=2):
    """Return a RGB tuple of the color.

    Parameters
    ----------
    color : None/tuple/string | None
        The color to use. Can either be None, or a tuple (R, G, B),
        a matplotlib color or an hexadecimal color '#...'.
    astype : type | np.float32
        The final color type.
    rmalpha : bool | True
        Specify if the alpha component have to be deleted.
    roundto : int | 2
        Number of digits per RGB.

    Returns
    -------
    coltuple: tuple
        Tuple of colors.
    """
    # Get the converted color :
    ccol = color2vb(color).ravel().astype(astype)
    # Round it :
    ccol = np.ndarray.tolist(np.around(ccol, roundto))
    if rmalpha:
        return tuple(ccol[0:-1])
    else:
        return tuple(ccol)


def array2colormap(x, cmap='inferno', clim=None, alpha=1.0, vmin=None,
                   vmax=None, under='dimgray', over='darkred',
                   translucent=None, faces_render=False):
    """Transform an array of data into colormap (array of RGBA).

    Parameters
    ----------
    x: array
        Array of data
    cmap : string | inferno
        Matplotlib colormap
    clim : tuple/list | None
        Limit of the colormap. The clim parameter must be a tuple / list
        of two float number each one describing respectively the (min, max)
        of the colormap. Every values under clim[0] or over clim[1] will
        peaked.
    alpha : float | 1.0
        The opacity to use. The alpha parameter must be between 0 and 1.
    vmin : float | None
        Threshold from which every color will have the color defined using
        the under parameter bellow.
    under : tuple/string | 'dimgray'
        Matplotlib color for values under vmin.
    vmax : float | None
        Threshold from which every color will have the color defined using
        the over parameter bellow.
    over : tuple/string | 'darkred'
        Matplotlib color for values over vmax.
    translucent : tuple | None
        Set a specific range translucent. With f_1 and f_2 two floats, if
        translucent is :

            * (f_1, f_2) : values between f_1 and f_2 are set to translucent
            * (None, f_2) x <= f_2 are set to translucent
            * (f_1, None) f_1 <= x are set to translucent
    faces_render : boll | False
        Precise if the render should be applied to faces

    Returns
    -------
    color : array_like
        Array of RGBA colors
    """
    # ================== Check input argument types ==================
    # Force data to be an array :
    x = np.asarray(x)

    # Check clim :
    clim = (None, None) if clim is None else list(clim)
    assert len(clim) == 2

    # ---------------------------
    # Check alpha :
    if (alpha < 0) or (alpha > 1):
        warn("The alpha parameter must be >= 0 and <= 1.")

    # ================== Define colormap ==================
    sc = cm.ScalarMappable(cmap=cmap)

    # Fix limits :
    norm = mplcol.Normalize(vmin=clim[0], vmax=clim[1])
    sc.set_norm(norm)

    # ================== Apply colormap ==================
    # Apply colormap to x :
    x_cmap = np.array(sc.to_rgba(x, alpha=alpha))

    # ================== Colormap (under, over) ==================
    if (vmin is not None) and (under is not None):
        under = color2vb(under)  # if isinstance(under, str) else under
        x_cmap[x < vmin, :] = under
    if (vmax is not None) and (over is not None):
        over = color2vb(over)  # if isinstance(over, str) else over
        x_cmap[x > vmax, :] = over

    # ================== Transparency ==================
    x_cmap = _transclucent_cmap(x, x_cmap, translucent)

    # Faces render (repeat the color to other dimensions):
    if faces_render:
        x_cmap = np.transpose(np.tile(x_cmap[..., np.newaxis],
                                      (1, 1, 3)), (0, 2, 1))

    return x_cmap.astype(np.float32)


def _transclucent_cmap(x, x_cmap, translucent, smooth=None):
    """Sub function to define transparency."""
    if translucent is not None:
        is_num = [isinstance(k, (int, float)) for k in translucent]
        assert len(translucent) == 2 and any(is_num)
        if all(is_num):                # (f_1, f_2)
            trans_x = np.logical_and(translucent[0] <= x, x <= translucent[1])
        elif is_num == [True, False]:  # (f_1, None)
            trans_x = translucent[0] <= x
        elif is_num == [False, True]:  # (None, f_2)
            trans_x = x <= translucent[1]
        x_cmap[..., -1] = np.invert(trans_x)
        if isinstance(smooth, int):
            alphas = x_cmap[:, -1]
            alphas = np.convolve(alphas, np.hanning(smooth), 'valid')
            alphas /= max(alphas.max(), 1.)
            x_cmap[smooth - 1::, -1] = alphas
    return x_cmap


def cmap_to_glsl(limits=None, lut_len=1024, color=None, **kwargs):
    """Get a glsl colormap.

    Parameters
    ----------
    limits : tuple | None
        Color limits for the object. Must be a tuple of two floats.
    lut_len : int | 1024
        Number of levels for the colormap.
    color : string | None
        Use a unique color for the colormap.
    kwarg : dict | None
        Additional inputs to pass to the array2colormap function.

    Returns
    -------
    cmap : vispy.color.Colormap
        VisPy colormap instance.
    """
    if limits is None:
        limits = (0., 1.)
    assert len(limits) == 2
    # Color transform :
    vec = np.linspace(limits[0], limits[1], lut_len)
    if color is None:  # colormap
        cmap = VispyColormap(array2colormap(vec, **kwargs))
    else:              # uniform color
        translucent = kwargs.get('translucent', None)
        rep_col = color2vb(color, length=lut_len)
        cmap_trans = _transclucent_cmap(vec, rep_col, translucent)
        cmap = VispyColormap(cmap_trans)

    return cmap


def dynamic_color(color, x, dynamic=(0., 1.)):
    """Dynamic color changing.

    Parameters
    ----------
    color : array_like
        The color to dynamic change. color must have a shape
        of (N, 4) RGBA colors
    x : array_like
        Dynamic values for color. x must have a shape of (N,)
    dynamic : tuple | (0.0, 1.0)
        Control the dynamic of color.

    Returns
    -------
    colordyn : array_like
        Dynamic color with a shape of (N, 4)
    """
    x = x.ravel()
    # Check inputs types :
    if color.shape[1] != 4:
        raise ValueError("Color must be RGBA")
    if color.shape[0] != len(x):
        raise ValueError("The length of color must be the same as"
                         " x: " + str(len(x)))
    # Normalise x :
    if dynamic[0] < dynamic[1]:
        x_norm = normalize(x, tomin=dynamic[0], tomax=dynamic[1])
    else:
        x_norm = np.full((len(x),), dynamic[0], dtype=np.float)
    # Update color :
    color[:, 3] = x_norm
    return color


def color2faces(color, length):
    """Pass a simple color to faces shape.

    Parameters
    ----------
    color : RGBA tuple
        Tuple of RGBA colors
    length : tuple
        Length of faces

    Returns
    -------
    color_face : array_like
        The color adapted for faces
    """
    color = np.asarray(color).ravel()
    colort = np.tile(np.array(color)[..., np.newaxis, np.newaxis],
                     (1, length, 3))
    return np.transpose(colort, (1, 2, 0))


def colorclip(x, th, kind='under'):
    """Force an array to have clipping values.

    Parameters
    ----------
    x : array_like
        Array of data.
    th : float
        The threshold to use.
    kind : string | 'under'
        Use eiher 'under' or 'over' for fore the array to clip for every
        values respectively under or over th.

    Returns
    -------
    x : array_like
        The clipping array.
    """
    if kind == 'under':
        idx = x < th
    elif kind == 'over':
        idx = x > th
    x[idx] = th
    return x


def type_coloring(color=None, n=1, data=None, rnd_dyn=(0.3, 0.9), clim=None,
                  cmap='viridis', vmin=None, under=None, vmax=None, over=None,
                  unicolor='gray'):
    """Switch between different coloring types.

    This function can be used to color a signal using random, uniform or
    dynamic colors.

    Parameters
    ----------
    color : string/tuple/array | None
        Choose how to color signals. Use None (or 'rnd', 'random') to
        generate random colors. Use 'uniform' (see the unicolor
        parameter) to define the same color for all signals. Use
        'dynamic' to have a dynamic color according to data values.
    n : int | 1
        The number of colors to generate in case of random or uniform
        colors.
    data : array_like | None
        The data to convert into color if the color type is dynamic.
        If this parameter is ignored, a default linear spaced vector of
        length n will be used instead.
    rnd_dyn : tuple | (.3, .9)
        Define the dynamic of random color. This parameter is active
        only if the color parameter is turned to None (or 'rnd' /
        'random').
    cmap : string | 'inferno'
        Matplotlib colormap (parameter active for 'dyn_minmax' and
        'dyn_time' color).
    clim : tuple/list | None
        Limit of the colormap. The clim parameter must be a tuple /
        list of two float number each one describing respectively the
        (min, max) of the colormap. Every values under clim[0] or over
        clim[1] will peaked (parameter active for 'dyn_minmax' and
        'dyn_time' color).
    alpha : float | 1.0
        The opacity to use. The alpha parameter must be between 0 and 1
        (parameter active for 'dyn_minmax' and 'dyn_time' color).
    vmin : float | None
        Threshold from which every color will have the color defined
        using the under parameter bellow (parameter active for
        'dyn_minmax' and 'dyn_time' color).
    under : tuple/string | 'gray'
        Matplotlib color for values under vmin (parameter active for
        'dyn_minmax' and 'dyn_time' color).
    vmax : float | None
        Threshold from which every color will have the color defined
        using the over parameter bellow (parameter active for
        'dyn_minmax' and 'dyn_time' color).
    over : tuple/string | 'red'
        Matplotlib color for values over vmax (parameter active for
        'dyn_minmax' and 'dyn_time' color).
    unicolor : tuple/string | 'gray'
        The color to use in case of uniform color.
    """
    # ---------------------------------------------------------------------
    # Random color :
    if color in [None, 'rnd', 'random']:
        # Create a (m, 3) color array :
        colout = np.random.uniform(size=(n, 3), low=rnd_dyn[0], high=rnd_dyn[1]
                                   )

    # ---------------------------------------------------------------------
    # Dynamic color :
    elif color == 'dynamic':
        # Generate a linearly spaced vector for None data :
        if data is None:
            data = np.arange(n)
        # Get colormap as (n, 3):
        colout = array2colormap(data.ravel(), cmap=cmap, clim=clim, vmin=vmin,
                                vmax=vmax, under=under, over=over)[:, 0:3]

    # ---------------------------------------------------------------------
    # Uniform color :
    elif color == 'uniform':
        # Create a (m, 3) color array :
        colout = color2vb(unicolor, length=n)[:, 0:3]

    # ---------------------------------------------------------------------
    # Not found color :
    else:
        raise ValueError("The color parameter is not recognized.")

    return colout.astype(np.float32)


def mpl_cmap(invert=False):
    """Get the list of matplotlib colormaps.

    Parameters
    ----------
    invert : bool | False
        Get the list of inverted colormaps.

    Returns
    -------
    cmap_lst: list
        list of available matplotlib colormaps.
    """
    # Full list of colormaps :
    fullmpl = list(cm.datad.keys()) + list(cm.cmaps_listed.keys())
    # Get the list of cmaps (inverted or not) :
    if invert:
        cmap_lst = [k for k in fullmpl if k.find('_r') + 1]
    else:
        cmap_lst = [k for k in fullmpl if not k.find('_r') + 1]

    # Sort the list :
    cmap_lst.sort()

    return cmap_lst


def mpl_cmap_index(cmap, cmaps=None):
    """Find the index of a colormap.

    Parameters
    ----------
    cmap : string
        Colormap name.
    cmaps : list | None
        List of colormaps.

    Returns
    -------
    idx : int
        Index of the colormap.
    invert : bool
        Boolean value indicating if it's a reversed colormap.
    """
    # Find if it's a reversed colormap :
    invert = bool(cmap.find('_r') + 1)
    # Get list of colormaps :
    if cmaps is None:
        cmap = cmap.replace('_r', '')
        cmaps = mpl_cmap()
        return np.where(np.char.find(cmaps, cmap) + 1)[0][0], invert
    else:
        return cmaps.index(cmap), invert



"""Wrappers."""
from functools import wraps


__all__ = ['wrap_properties']


def wrap_properties(fn):
    """Run properties if not None."""
    @wraps(fn)
    def wrapper(self, value):
        if value is not None:
            fn(self, value)
    return wrapper



"""Pacmap object."""
import numpy as np

from .image_obj import ImageObj
from ..io.dependencies import is_tensorpac_installed


class PacmapObj(ImageObj):
    """Create a Phase-Amplitude Coupling (PAC) object.

    The PAC is computed using the tensorpac package.

    The Pacmap can be used to visualize :

        * PAC, across time for a fixed phase and several amplitude frequencies
        * PAC, across time for a fixed amplitude and several phase frequencies
        * PAC is computed across time for several amplitudes and phase.

    Parameters
    ----------
    name : string | None
        Name of the pacmap object.
    data : array_like
        Array of data of shape (N,)
    sf : float | 1.
        The sampling frequency.
    f_pha : list | [(2, 4), (5, 7), (8, 13)]
        The phase vector.
    f_amp : list | [(40, 60), (60, 100)]
        The amplitude vector.
    idpac : tuple | (4, 0, 0)
        The PAC method to use.
    n_window : int | None
        Number of time points to consider when computing pac for a fixed phase
        or amplitude.
    clim : tuple | None
        Colorbar limits. If None, `clim=(data.min(), data.max())`
    cmap : string | None
        Colormap name.
    vmin : float | None
        Minimum threshold of the colorbar.
    under : string/tuple/array_like | None
        Color for values under vmin.
    vmax : float | None
        Maximum threshold of the colorbar.
    under : string/tuple/array_like | None
        Color for values over vmax.
    interpolation : string | 'nearest'
        Interpolation method for the image. See vispy.scene.visuals.Image for
        availables interpolation methods.
    max_pts : int | -1
        Maximum number of points of the image along the x or y axis. This
        parameter is essentially used to solve OpenGL issues with very large
        images.
    transform : VisPy.visuals.transforms | None
        VisPy transformation to set to the parent node.
    parent : VisPy.parent | None
        Markers object parent.
    verbose : string
        Verbosity level.
    kw : dict | {}
        Optional arguments are used to control the colorbar
        (See :class:`ColorbarObj`).

    Notes
    -----
    List of supported shortcuts :

        * **s** : save the figure
        * **<delete>** : reset camera

    Examples
    --------
    >>> import numpy as np
    >>> from sleep_scoring.objects import SpectrogramObj
    >>> n, sf = 512, 256  # number of time-points and sampling frequency
    >>> time = np.arange(n) / sf  # time vector
    >>> data = np.sin(2 * np.pi * 25. * time) + np.random.rand(n)
    >>> spec = SpectrogramObj('spec', data, sf)
    >>> spec.preview(axis=True)
    """

    def __init__(self, name, data=None, sf=1., f_pha=[(2, 4), (5, 7), (8, 13)],
                 f_amp=[(40, 60), (60, 100)], idpac=(4, 0, 0), n_window=None,
                 cmap='viridis', clim=None, vmin=None, under='gray', vmax=None,
                 over='red', interpolation='nearest', max_pts=-1, parent=None,
                 transform=None, verbose=None, pac_kw={}, **kw):
        """Init."""
        # Initialize the image object :
        ImageObj.__init__(self, name, interpolation=interpolation,
                          max_pts=max_pts, parent=parent, transform=transform,
                          verbose=verbose, **kw)

        # Compute pacmap and set data to the ImageObj :
        if isinstance(data, np.ndarray):
            self.set_data(data, sf, f_pha, f_amp, idpac, n_window, clim, cmap,
                          vmin, under, vmax, over, **pac_kw)

    def set_data(self, data, sf=1., f_pha=[(2, 4), (5, 7), (8, 13)],
                 f_amp=[(40, 60), (60, 100)], idpac=(4, 0, 0), n_window=None,
                 clim=None, cmap=None, vmin=None, under=None, vmax=None,
                 over=None, **kwargs):
        """Compute pacmap and set data to the ImageObj."""
        # ======================= CHECKING =======================
        is_tensorpac_installed(raise_error=True)
        from tensorpac import Pac
        data = np.squeeze(data)
        assert isinstance(data, np.ndarray) and data.ndim == 1
        assert isinstance(sf, (int, float))
        time = np.arange(len(data))
        # Switch between several plotting capabilities :
        _pha = len(f_pha) == 2 and all([isinstance(k, (int,
                                                       float)) for k in f_pha])
        _amp = len(f_amp) == 2 and all([isinstance(k, (int,
                                                       float)) for k in f_amp])
        if _pha or _amp:
            assert isinstance(n_window, int)
            sections = time.copy()[::n_window][1::]
            time = np.array(np.array_split(time, sections)[0:-1]).mean(1).T
            time /= sf
            data = np.array(np.array_split(data, sections)[0:-1]).T
        # Define the pac object and compute pac :
        p = Pac(idpac=idpac, fpha=f_pha, famp=f_amp, **kwargs)
        pac = np.squeeze(p.filterfit(sf, data, njobs=1, axis=0))
        pac[np.isnan(pac)] = 0.
        assert pac.ndim == 2
        # Get x-axis and y-axis :
        if _pha:
            xaxis, yaxis = time, p.yvec
        elif _amp:
            xaxis, yaxis = time, p.xvec
        else:
            xaxis, yaxis = p.xvec, p.yvec
        # Update color arguments :
        self._update_cbar_args(cmap, clim, vmin, vmax, under, over)

        # Set data to the image object :
        ImageObj.set_data(self, pac, xaxis=xaxis, yaxis=yaxis,
                          **self.to_kwargs())



"""Create basic popup."""

from PyQt5.Qt import QWidget, QRect
from PyQt5 import QtWidgets, QtCore
import webbrowser

from .screenshot_gui import Ui_Screenshot
from ..guitools import textline2color

__all__ = ('ShortcutPopup', 'ScreenshotPopup', 'HelpMenu')


class ShortcutPopup(QWidget):
    """Popup window with the list of shorcuts."""

    def __init__(self):
        """Init."""
        QWidget.__init__(self)
        self.setGeometry(QRect(400, 200, 700, 600))
        layout = QtWidgets.QGridLayout(self)
        self.table = QtWidgets.QTableWidget()
        self.table.setColumnCount(2)
        self.table.horizontalHeader().setStretchLastSection(True)
        self.table.verticalHeader().setStretchLastSection(False)
        layout.addWidget(self.table)
        # Add column names :
        item = QtWidgets.QTableWidgetItem()
        self.table.setHorizontalHeaderItem(0, item)
        item = self.table.horizontalHeaderItem(0)
        item.setText("Keys")
        self.table.setHorizontalHeaderItem(1, item)
        item = self.table.horizontalHeaderItem(1)
        item.setText("Description")

    def set_shortcuts(self, shdic):
        """Fill table."""
        self.table.setRowCount(len(shdic))
        for num, k in enumerate(shdic):
            self.table.setItem(num, 0, QtWidgets.QTableWidgetItem(k[0]))
            self.table.setItem(num, 1, QtWidgets.QTableWidgetItem(k[1]))


class ScreenshotPopup(Ui_Screenshot):
    """Popup window with the list of shorcuts."""

    def __init__(self, fcn_run, canvas_names=[]):
        """Init."""
        super(ScreenshotPopup, self).__init__()
        self._ss = QtWidgets.QWidget()
        self.setupUi(self._ss)

        # Render entire window / selected canvas :
        self._ssSelect.currentIndexChanged.connect(self._fcn_select_render)
        # Add canvas names :
        self._ssCanvasName.addItems(canvas_names)
        # Image resolution :
        self._ssResolution.currentIndexChanged.connect(self._fcn_resolution)
        # Background color :
        self._ssBgcolorChk.clicked.connect(self._fcn_enable_bgcolor)
        # Run :
        self._ssRun.clicked.connect(fcn_run)

        self._fcn_select_render()
        self._fcn_resolution()

    def _fcn_select_render(self):
        """Render either the entire window / selected canvas."""
        txt = int(self._ssSelect.currentIndex())
        self._ssCanvasW.setVisible(txt == 0)

    def _fcn_resolution(self):
        """Choose resolution method."""
        idx = int(self._ssResolution.currentIndex())
        if idx == 0:  # custom
            self._ssCustomW.setVisible(True)
            self._ssFactorW.setVisible(False)
        elif idx == 1:  # factor
            self._ssCustomW.setVisible(False)
            self._ssFactorW.setVisible(True)

    def _fcn_enable_bgcolor(self):
        """Enable background color."""
        self._ssBgcolor.setEnabled(self._ssBgcolorChk.isChecked())

    def show(self):
        """Show widget."""
        self._ss.show()

    def to_kwargs(self):
        """Get arguments."""
        kwargs = {}
        kwargs['entire'] = int(self._ssSelect.currentIndex()) == 1
        kwargs['canvas'] = str(self._ssCanvasName.currentText())
        resolution = int(self._ssResolution.currentIndex())
        if resolution == 0:
            kwargs['print_size'] = (self._ssWidth.value(),
                                    self._ssHeight.value())
        else:
            kwargs['print_size'] = None
        kwargs['unit'] = str(self._ssUnit.currentText())
        kwargs['dpi'] = self._ssDpi.value()
        kwargs['factor'] = self._ssFactor.value()
        kwargs['autocrop'] = self._ssAutocrop.isChecked()
        if self._ssBgcolorChk.isChecked():
            bgcolor = textline2color(str(self._ssBgcolor.text()))[1]
            kwargs['bgcolor'] = bgcolor
        else:
            kwargs['bgcolor'] = None
        kwargs['transparent'] = self._ssTransp.isChecked()
        return kwargs


class HelpMenu(object):
    """Help menu creation class.

    This class create a help menu with a link to the documentation and a link
    to download the pdf version of the documentation.
    """

    def __init__(self, doc_section, add_shortcuts=True):
        """Init."""
        _translate = QtCore.QCoreApplication.translate
        # Main Help menu creation :
        help_menu = QtWidgets.QMenu('Help', self)
        self.menuBar().addMenu(help_menu)
        if add_shortcuts:
            # Shortcuts popup window :
            self._shpopup = ShortcutPopup()
            # Shortcuts :
            shortcuts = QtWidgets.QAction("&Shortcuts", self)
            shortcuts.setShortcut(_translate("MainWindow", "Ctrl+T"))
            shortcuts.triggered.connect(self._shpopup.show)
            help_menu.addAction(shortcuts)
        # Help section submenu :
        help_section = QtWidgets.QMenu('Open help section about', self)
        help_menu.addMenu(help_section)

        for menu, url in doc_section.items():
            # Create action :
            section_action = QtWidgets.QAction("&" + menu, self)
            # Action function :

            def define(url):
                def _fcn_open_help_section():
                    """Open the help section."""
                    webbrowser.open(url)
                return _fcn_open_help_section
            # Triggered function :
            section_action.triggered.connect(define(url))
            # Add action to the menu :
            help_section.addAction(section_action)
        # PDF documentation action :
        pdf_doc = QtWidgets.QAction("&Download doc (pdf)", self)
        pdf_doc.triggered.connect(self._fcn_open_pdf_doc)
        help_menu.addAction(pdf_doc)

    def _fcn_open_pdf_doc(self):
        """Open the pdf version of the documentation."""
        webbrowser.open("https://drive.google.com/open?id=0B6vt"
                        "JiCQZUBvdVphQWw1RFBSZW8")



"""Visbrain configurations."""
import sys
import getopt
import logging

from PyQt5 import QtWidgets
from vispy import app as visapp

from .utils import Profiler, set_log_level

"""Set 'info' as the default logging level
"""
logger = logging.getLogger('visbrain')
set_log_level('info')

"""Configuration dict
"""
CONFIG = {}

"""Visbrain profiler (derived from the VisPy profiler)
"""
PROFILER = Profiler()

"""PyQt application
"""
PYQT_APP = QtWidgets.QApplication.instance()
if PYQT_APP is None:
    PYQT_APP = QtWidgets.QApplication([''])
CONFIG['PYQT_APP'] = PYQT_APP
CONFIG['SHOW_PYQT_APP'] = True

"""VisPy application
"""
CONFIG['VISPY_APP'] = visapp.application.Application()


def use_app(backend_name):
    """Use a specific backend."""
    CONFIG['VISPY_APP'] = visapp.application.Application(backend_name)


"""Input command line arguments
"""
VISBRAIN_HELP = """
Visbrain command line arguments:

  --visbrain-log=(profiler|debug|info|warning|error|critical)
    Sets the verbosity of logging output. The default is 'info'.

  --visbrain-search=[search string]
    Search string in logs.

  --visbrain-show=(True|False)
    Control if GUI have to be displayed.

  --visbrain-help
    Display help Visbrain command line help.

"""


def init_config(argv):
    """Initialize visbrain configuration."""
    global CONFIG

    argnames = ['visbrain-log=', 'visbrain-show=', 'visbrain-help',
                'visbrain-search=']
    try:
        opts, args = getopt.getopt(sys.argv[1:], '', argnames)
    except getopt.GetoptError:
        opts = []

    for o, a in opts:
        if o.startswith('--visbrain'):
            if o == '--visbrain-help':
                print(VISBRAIN_HELP)
            if o == '--visbrain-log':
                set_log_level(a)
            if o == '--visbrain-show':
                CONFIG['SHOW_PYQT_APP'] = eval(a)
                logger.debug("Show PyQt app : %r" % CONFIG['SHOW_PYQT_APP'])
            if o == '--visbrain-search':
                set_log_level(match=a)

init_config(sys.argv[1:])  # noqa



"""Load read and write functions."""
from .dependencies import *  # noqa
from .dialog import *  # noqa
from .download import *  # noqa
from .mneio import *  # noqa
from .path import *  # noqa
from .read_annotations import *  # noqa
from .read_data import *  # noqa
from .rw_nifti import (read_nifti, read_mist, niimg_to_transform)  # noqa
from .read_sleep import *  # noqa
from .rw_config import *  # noqa
from .rw_hypno import *  # noqa
from .rw_utils import *  # noqa
from .write_data import *  # noqa
from .write_image import *  # noqa
from .write_table import *  # noqa
from .write_template import *  # noqa



"""Manage colorbar of multiple objects."""
from .CbarBase import CbarBase
from ...io import save_config_json, load_config_json

__all__ = ['CbarObjetcs']


class CbarObjetcs(object):
    """Create and manage multiple colorbar objects."""

    def __init__(self):
        """Init."""
        self._objs = {}
        self._selected = None

    def __getitem__(self, key):
        """Get item in the curently selected colorbar object."""
        return self._objs[self._selected][key]

    def __setitem__(self, key, value):
        """Set item in the curently selected colorbar object."""
        exec("self._objs[self._selected]._" + key + " = value")

    def __iter__(self):
        """Iterate over the selected colorbar object."""
        for key, val in self._objs[self._selected].to_dict().items():
            yield key, val

    def __bool__(self):
        """Test if it's not empty."""
        return self._objs and self._selected

    def select(self, name):
        """Select an object.

        Parameters
        ----------
        name : string
            Object name.
        """
        if name in self._objs.keys():
            self._selected = name
        else:
            raise ValueError(name + " not in the object list.")

    def add_object(self, name, obj, overwrite=True):
        """Add a colorbar object.

        Parameters
        ----------
        name : string
            Name of the object.
        obj : CbarBase
            The CbarBase object to add.
        overwrite : bool | True
            If a colorbar object already has the same name, define if it
            should be overwritten.
        """
        # Test if name is a string :
        if not isinstance(name, str):
            raise ValueError("name must be a string.")
        # Test if the name is free :
        if not overwrite and (name in self._objs.keys()):
            raise ValueError("The colorbar object '" + name + "' already "
                             "exist. Use an other name")
        if isinstance(obj, CbarBase):
            self._selected = name
            self._objs[name] = obj
        else:
            raise ValueError("The object to add must be a CbarBase type.")

    def to_kwargs(self):
        """Return a dictionary for input arguments."""
        return self._objs[self._selected].to_kwargs()

    def to_dict(self, alldicts=False):
        """Return a dictionary of all colorbar args."""
        if alldicts:
            return {k: i.to_dict() for k, i in self._objs.items()}
        else:
            return self._objs[self._selected].to_dict()

    def keys(self):
        """Return the list of entries in the object dict."""
        return list(self._objs.keys())

    def save(self, filename):
        """Save all colorbar configurations.

        Parameters
        ----------
        filename : string
            Name of the file to be saved.
        """
        # Build the configuration file :
        config = {k: i.to_dict() for k, i in self._objs.items()}
        # Save in a txt file :
        save_config_json(filename, config)

    def load(self, filename, clean=True, select=None):
        """Load a colorbar configuration file.

        Parameters
        ----------
        filename: string
            Name of the file to load.
        clean : bool | True
            Specify if all objects have to be clean before.
        select : string | None
            Name of the object to select on load.
        """
        # Load the configuration :
        config = load_config_json(filename)
        # Clean current objects :
        if clean:
            self._objs = {}
        # Load objects :
        for k, i in config.items():
            self.add_object(k, CbarBase(**i))
        # Select the first one :
        if (select is None) or (select not in self._objs.keys()):
            self._selected = list(self._objs.keys())[0]

    def from_dict(self, dico, clean=True):
        """Define cbar objects from dictionary."""
        # Clean :
        if clean:
            self._objs = {}
        # Load objects :
        for k, i in dico.items():
            self._objs[k] = CbarBase(**i)

    def update(self):
        """Update function when an attribute change."""
        self._objs[self._selected]._fcn()

    def autoscale(self):
        """Execute autoscale function associated to the object."""
        self._objs[self._selected]._minmaxfcn()

    # ----------- NAME -----------
    @property
    def name(self):
        """Get the name value."""
        return self._selected



"""Most basic colorbar class."""
from ...utils import color2tuple, wrap_properties, cmap_to_glsl

__all__ = ('CbarArgs', 'CbarBase')


class CbarArgs(object):
    """Manage the different inputs for the colormap creation.

    Parameters
    ----------
    cmap : string | None
        Matplotlib colormap (like 'viridis', 'inferno'...).
    clim : tuple/list | None
        Colorbar limit. Every values under / over clim will
        clip.
    isvmin : bool | False
        Activate/deactivate vmin.
    vmin : float | None
        Every values under vmin will have the color defined
        using the under parameter.
    isvmax : bool | False
        Activate/deactivate vmax.
    vmax : float | None
        Every values over vmin will have the color defined
        using the over parameter.
    under : tuple/string | None
        Matplotlib color under vmin.
    over : tuple/string | None
        Matplotlib color over vmax.
    """

    def __init__(self, cmap=None, clim=None, isvmin=False, vmin=None,
                 isvmax=False, vmax=None, under=None, over=None):
        """Init."""
        # Cmap/Clim/Vmin/Vmax/Under/Over :
        self._cmap, self._clim = cmap, clim
        self._vmin, self._vmax = vmin, vmax
        self._isvmin, self._isvmax = isvmin, isvmax
        self._under, self._over = under, over
        self._minmax = None

    def to_kwargs(self, addisminmax=False):
        """Return a dictionary for input arguments.

        Parameters
        ----------
        addisminmax : bool | False
            Specify if the returned dictionary have to had the ismin and
            ismax variables.
        """
        kwargs = {}
        kwargs['cmap'] = self._cmap
        kwargs['clim'] = self._clim
        kwargs['vmin'] = self._vmin if self._isvmin else None
        kwargs['under'] = self._under
        kwargs['vmax'] = self._vmax if self._isvmax else None
        kwargs['over'] = self._over
        if addisminmax:
            kwargs['isvmin'], kwargs['isvmax'] = self._isvmin, self._isvmax
        return kwargs

    def _get_glsl_colormap(self, limits=None):
        """Get a GLSL version of the colormap.

        Parameters
        ----------
        limits : tuple | None
            A tuple of floats defining the limits of the data.

        Returns
        -------
        cmap : class:`vispy.color.Colormap`
            Colormap instance.
        """
        limits = limits if limits is not None else self._minmax
        return cmap_to_glsl(limits=limits, **self.to_kwargs())

    def _update_cbar_args(self, cmap, clim, vmin, vmax, under, over):
        """Update colorbar elements."""
        kw = dict(clim=clim, cmap=cmap, vmin=vmin, vmax=vmax, under=under,
                  over=over)
        self._isvmax = isinstance(vmax, (int, float))
        self._isvmin = isinstance(vmin, (int, float))
        self.update_from_dict(kw)
        return kw

    def update_from_dict(self, kwargs):
        """Update attributes from a dictionary."""
        for k, i in kwargs.items():
            off = '_' if k != 'cmap' else ''
            if not isinstance(i, str):
                exec('self.' + off + k + '=' + str(i))
            else:
                exec("self." + off + k + "='" + str(i) + "'")

    # ----------- CMAP -----------
    @property
    def cmap(self):
        """Get the cmap value."""
        return self._cmap

    @cmap.setter
    @wrap_properties
    def cmap(self, value):
        """Set cmap value."""
        self._cmap = value


class CbarBase(CbarArgs):
    """Base class for colorbar.

    Parameters
    ----------
    cblabel : string | ''
        Colorbar label.
    cbtxtsz : float | 5..
        Text size of the colorbar label.
    cbtxtsh : float | 2.3
        Shift for the colorbar label.
    txtcolor : string | 'white'
        Text color.
    txtsz : float | 3.
        Text size for clim/vmin/vmax text.
    txtsh : float | 1.2
        Shift for clim/vmin/vmax text.
    border : bool | True
        Display colorbar borders.
    bw : float | 2.
        Border width.
    limtxt : bool | True
        Display vmin/vmax text.
    bgcolor : tuple/string | (0., 0., 0.)
        Background color of the colorbar canvas.
    ndigits : int | 2
        Number of digits for the text.
    """

    def __init__(self, cmap='viridis', clim=(0, 1), vmin=None, isvmin=False,
                 vmax=None, isvmax=False, under='gray', over='red', cblabel='',
                 cbtxtsz=5., cbtxtsh=2.3, txtcolor='white', txtsz=3.,
                 txtsh=1.2, width=.17, border=True, bw=2., limtxt=True,
                 bgcolor=(0., 0., 0.), ndigits=2, minmax=None, fcn=None,
                 minmaxfcn=None):
        """Init."""
        CbarArgs.__init__(self, cmap, clim, isvmin, vmin, isvmax, vmax, under,
                          over)
        # Cb text :
        self._cblabel = cblabel
        self._cbtxtsz = cbtxtsz
        self._cbtxtsh = cbtxtsh
        # Text :
        self._txtcolor = txtcolor
        self._txtsz = txtsz
        self._txtsh = txtsh
        self._limtxt = limtxt
        # Settings :
        self._bgcolor = bgcolor
        self._border = border
        self._bw = bw
        self._ndigits = ndigits
        self._width = width
        if fcn is None:
            def fcn():
                pass
        self._fcn = fcn
        if minmaxfcn is None:
            def minmaxfcn():
                pass
        self._minmaxfcn = minmaxfcn

    def __getitem__(self, key):
        """Get item (usefull for CbarObjects)."""
        return eval('self._' + key)

    def __setitem__(self, key, value):
        """Set self items."""
        if not isinstance(value, str):
            exec("self._" + key + "=" + str(value))
        else:
            exec("self._" + key + "='" + value + "'")

    # -------------------------------------------------------------------------
    #                             USER METHODS
    # -------------------------------------------------------------------------
    def to_dict(self):
        """Return a dictionary of all colorbar args."""
        cblab = self._cblabel
        if not cblab and hasattr(self, '_default_cblabel'):
            cblab = self._default_cblabel
        if self._clim is None:
            self._clim = (0., 1.)
        to = dict(cmap=self._cmap, clim=[float(k) for k in self._clim],
                  isvmin=self._isvmin, vmin=self._vmin, vmax=self._vmax,
                  under=list(color2tuple(self._under, float)),
                  over=list(color2tuple(self._over, float)),
                  isvmax=self._isvmax, cblabel=cblab,
                  cbtxtsz=float(self._cbtxtsz), cbtxtsh=float(self._cbtxtsh),
                  txtcolor=list(color2tuple(self._txtcolor, float)),
                  txtsz=float(self._txtsz), txtsh=float(self._txtsh),
                  border=self._border, bw=float(self._bw), limtxt=self._limtxt,
                  bgcolor=list(color2tuple(self._bgcolor, float)),
                  ndigits=int(self._ndigits), width=float(self._width))
        return to

    def update(self):
        """Fonction to run when an update is needed."""
        if self._fcn is not None:
            self._fcn()
        else:
            raise ValueError("No updating function found.")



"""Hypnogram object."""
import os

import numpy as np
from vispy import scene

from .visbrain_obj import VisbrainObject
from ..visuals import Hypnogram
from ..io import read_hypno


class HypnogramObj(VisbrainObject):
    """Hypnogram object.

    Parameters
    ----------
    name : string
        Name of the hypnogram object or path to a *.txt or *.csv file.
    data : array_like
        Array of data of shape (n_pts,).
    time : array_like | None
        Array of time points of shape (n_pts,)
    datafile : string | None
        Path to the data file.
    art, wake, rem, n1, n2, n3 :
        Stage identification inside the data array.
    art_visual, wake_visual, rem_visual, n1_visual, n2_visual, n3_visual :
        Stage order when plotting.
    art_color, wake_color, rem_color, n1_color, n2_color, n3_color :
        Stage color.
    line_width : float | 2.
        Line with of the hypnogram.
    antialias : bool | False
        Use anti-aliasing line.
    unicolor : bool | False
        Use a uni black color for the hypnogram.
    transform : VisPy.visuals.transforms | None
        VisPy transformation to set to the parent node.
    parent : VisPy.parent | None
        Hypnogram object parent.
    verbose : string
        Verbosity level.
    kw : dict | {}
        Optional arguments are used to control the colorbar
        (See :class:`ColorbarObj`).

    Notes
    -----
    List of supported shortcuts :

        * **s** : save the figure
        * **<delete>** : reset camera

    Examples
    --------
    >>> import numpy as np
    >>> from sleep_scoring.objects import HypnogramObj
    >>> data = np.repeat(np.arange(6), 100) - 1.
    >>> h_obj = HypnogramObj('hypno', data)
    >>> h_obj.preview(axis=True)
    """

    def __init__(self, name, data=None, time=None, datafile=None, art=-1,
                 wake=0, n1=1, n2=2, n3=3, rem=4, art_visual=1, wake_visual=0,
                 rem_visual=-1, n1_visual=-2, n2_visual=-3, n3_visual=-4,
                 art_color='#8bbf56', wake_color='#56bf8b',
                 rem_color='#bf5656', n1_color='#aabcce', n2_color='#405c79',
                 n3_color='#0b1c2c', line_width=2., antialias=False,
                 unicolor=False, transform=None, parent=None, verbose=None,
                 **kw):
        """Init."""
        # Load *.txt, *.csv and *.hyp files :
        file, ext = os.path.splitext(name)
        if ext in ['.csv', '.txt', '.hyp', '.xlsx', '.edf']:
            if (ext == '.xlsx') and (time is None):
                raise ValueError("The `time` input should not be None with "
                                 "excel files. Use a NumPy array instead.")
            data, sf = read_hypno(name, time=time, datafile=datafile)
            name, time = os.path.split(name)[1], np.arange(len(data)) / sf
        # Initialize VisbrainObject and Hypnogram visuam creation :
        VisbrainObject.__init__(self, name, parent, transform, verbose, **kw)
        self.line = Hypnogram(data, time, art, wake, n1, n2, n3, rem,
                              art_visual, wake_visual, rem_visual, n1_visual,
                              n2_visual, n3_visual, art_color, wake_color,
                              rem_color, n1_color, n2_color, n3_color,
                              line_width, antialias, unicolor,
                              parent=self._node)

    def _get_camera(self):
        t_min, t_max = self.line.time.min(), self.line.time.max()
        d_min, d_max = self.line.min_visual(), self.line.max_visual()
        rect = (t_min, d_min - .5, t_max - t_min, d_max - d_min + 1.)
        return scene.cameras.PanZoomCamera(rect=rect)

    def set_stage(self, stage, idx_start, idx_end):
        """Set stage.

        Parameters
        ----------
        stage : str, int
            Stage to define. Should either be a string (e.g 'art', 'rem'...) or
            an integer.
        idx_start : int
            Index where the stage begin.
        idx_end : int
            Index where the stage finish.
        """
        self.line.set_stage(stage, idx_start, idx_end)

    # ----------- LINE_WIDTH -----------
    @property
    def line_width(self):
        """Get the line_width value."""
        return self.line._line_width

    @line_width.setter
    def line_width(self, value):
        """Set line_width value."""
        self.line.line_width = value

    # ----------- UNICOLOR -----------
    @property
    def unicolor(self):
        """Get the unicolor value."""
        return self.line._unicolor

    @unicolor.setter
    def unicolor(self, value):
        """Set unicolor value."""
        self.line.unicolor = value



"""Set of functions for picture managment."""
import numpy as np
# Removed as of scipy=1.3.0
# from scipy.misc import imresize
from PIL import Image


__all__ = ('piccrop', 'picresize')


def piccrop(im, margin=10):
    """Automatic picture cropping.

    Parameters
    ----------
    im : array_like
        The array of image data. Could be a (N, M) or (N, M, 3/4).
    margin : int | 10
        Number of pixels before and after to condider for cropping
        security.

    Returns
    -------
    imas : array_like
        The cropped figure.
    """
    # ================= Size checking =================
    if im.ndim < 2:
        raise ValueError("im must have at least two dimensions.")
    elif im.ndim == 3:
        imas = im[..., 0:3].mean(axis=2)
    else:
        imas = im

    # ================= Derivative =================
    imdiff_x = np.diff(imas, axis=1)
    imdiff_y = np.diff(imas, axis=0)

    # ================= Cropping start / finish =================
    # x-axis :
    idx_x = np.where(imdiff_x != 0)[1]
    if idx_x.size:
        ncols = imas.shape[1]
        x_min = max(0, idx_x.min() - margin + 1)
        x_max = min(min(ncols, idx_x.max() + margin + 1), ncols)
        sl_x = slice(x_min, x_max)
    else:
        sl_x = slice(None)
    # y-axis :
    idx_y = np.where(imdiff_y)[0]
    if idx_y.size:
        y_min = max(0, idx_y.min() - margin + 1)
        nrows = imas.shape[0]
        y_max = min(min(imas.shape[0], idx_y.max() + margin + 1), nrows)
        sl_y = slice(y_min, y_max)
    else:
        sl_y = slice(None)

    return im[sl_y, sl_x, ...]


def picresize(im, axis=0, extend=False):
    """For a list of pictures, resize them all to the same size.

    Inspect each picture in the list, get all shapes and use the smallest or
    the largest picture as the reference for resizing all other pictures.

    Parameters
    ----------
    im : list
        List of np.ndarray of shapes (N, M) or (N, M, 3/4)
    axis : int | 0
        Specify which axis is considered as the reference. Use 0 and all
        figures will have the same height otherwise use 1 for width.
    extend : bool | False
        Specify if the reference picture have to be the smallest
        (False - downsize all pictures) or the largest
        (True - extend all others).

    Returns
    -------
    imr : list
        List of resized pictures.
    """
    # ================= Checking =================
    if not isinstance(im, list):
        raise ValueError("im must be a list of pictures.")
    if axis not in [0, 1]:
        raise ValueError("axis parameter must either be 0 or 1.")

    # ================= Shapes =================
    sh = np.array([float(k.shape[axis]) for k in im])
    factors = sh.max() / sh if extend else sh.min() / sh

    # ================= Resize =================
    # return [imresize(k, i) for k, i in zip(im, factors)]
    # TODO: not sure if this syntax is right
    return [np.array(Image.fromarray(k).resize(i)) for k, i in zip(im, factors)]



"""Main class for settings managment."""
from PyQt5 import QtCore, QtGui, QtWidgets

import numpy as np

from ..ui_init import AxisCanvas, TimeAxis
from sleep_scoring.utils import mpl_cmap, color2vb
from sleep_scoring.config import PROFILER
from sleep_scoring.io.dependencies import is_lspopt_installed

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):  # noqa
        """From utf8 pyqt function."""
        return s


class UiPanels(object):
    """Main class for settings managment."""

    def __init__(self):
        """Init."""
        # Bold font :
        self._font = QtGui.QFont()
        self._font.setBold(True)
        self._addspace = '   '
        self._pan_pick.currentIndexChanged.connect(self._fcn_pan_pick)

        # =====================================================================
        # MAIN GRID :
        # =====================================================================
        self._chanGrid = QtWidgets.QGridLayout()
        self._chanGrid.setContentsMargins(-1, -1, -1, 6)
        self._chanGrid.setSpacing(2)
        self._chanGrid.setObjectName(_fromUtf8("_chanGrid"))
        self.gridLayout_21.addLayout(self._chanGrid, 0, 0, 1, 1)
        PROFILER("Channel grid", level=2)

        # =====================================================================
        # CHANNELS
        # =====================================================================
        # Create check buttons and panels for every channel :
        self._fcn_chan_check_and_create_w()
        self._PanChanSelectAll.clicked.connect(self._fcn_select_all_chan)
        self._PanChanDeselectAll.clicked.connect(self._fcn_deselect_all_chan)
        self._PanAmpAuto.clicked.connect(self._fcn_chan_auto_amp)
        self._PanAmpSym.clicked.connect(self._fcn_chan_sym_amp)
        self._channels_lw.valueChanged.connect(self._fcn_chan_set_lw)
        self._channels_alias.clicked.connect(self._fcn_chan_antialias)
        PROFILER("Channel canvas, widgets and buttons", level=2)

        # =====================================================================
        # AMPLITUDES
        # =====================================================================
        # Save all current amplitudes :
        self._PanAllAmpMax.setValue(100.)
        self._ylims = np.zeros((len(self), 2), dtype=np.float32)
        self._fcn_update_amp_info()
        self._PanAllAmpMin.valueChanged.connect(self._fcn_all_amp)
        self._PanAllAmpMax.valueChanged.connect(self._fcn_all_amp)
        PROFILER("Channel amplitudes", level=2)

        # =====================================================================
        # SPECTROGRAM
        # =====================================================================
        # Main canvas for the spectrogram :
        self._specCanvas = AxisCanvas(axis=self._ax, name='Spectrogram',
                                      fcn=[self.on_mouse_wheel])
        self._SpecW, self._SpecLayout = self._create_compatible_w("SpecW",
                                                                  "SpecL")
        self._SpecLayout.addWidget(self._specCanvas.canvas.native)
        self._chanGrid.addWidget(self._SpecW, len(self) + 1, 1, 1, 1)
        # Add label :
        self._specLabel = QtWidgets.QLabel(self.centralwidget)
        self._specLabel.setText(self._addspace + self._channels[0])
        self._specLabel.setFont(self._font)
        self._chanGrid.addWidget(self._specLabel, len(self) + 1, 0, 1, 1)
        # Add list of colormaps :
        self._cmap_lst = mpl_cmap()
        self._PanSpecCmap.addItems(self._cmap_lst)
        self._PanSpecCmap.setCurrentIndex(self._cmap_lst.index(self._defcmap))
        # Add list of channels :
        self._PanSpecChan.addItems(self._channels)
        # Disable multitaper option if not is_lspopt_installed
        self._PanSpecMethod.model().item(2).setEnabled(is_lspopt_installed())
        # Connect spectrogam properties :
        self._PanSpecApply.setEnabled(False)
        self._PanSpecApply.clicked.connect(self._fcn_spec_set_data)
        self._PanSpecNfft.valueChanged.connect(self._fcn_spec_compat)
        self._PanSpecStep.valueChanged.connect(self._fcn_spec_compat)
        self._PanSpecFstart.valueChanged.connect(self._fcn_spec_compat)
        self._PanSpecFend.valueChanged.connect(self._fcn_spec_compat)
        self._PanSpecCon.valueChanged.connect(self._fcn_spec_compat)
        self._PanSpecCmap.currentIndexChanged.connect(self._fcn_spec_compat)
        self._PanSpecChan.currentIndexChanged.connect(self._fcn_spec_compat)
        self._PanSpecMethod.currentIndexChanged.connect(self._fcn_spec_compat)
        self._PanSpecCmapInv.clicked.connect(self._fcn_spec_compat)
        self._PanSpecNorm.currentIndexChanged.connect(self._fcn_spec_compat)
        self._PanSpecInterp.currentIndexChanged.connect(self._fcn_spec_interp)
        PROFILER("Spectrogram", level=2)

        # =====================================================================
        # HYPNOGRAM
        # =====================================================================
        self._hypCanvas = AxisCanvas(axis=self._ax, name='Hypnogram',
                                     fcn=[self.on_mouse_wheel], use_pad=True)
        self._HypW, self._HypLayout = self._create_compatible_w("HypW", "HypL")
        self._HypLayout.addWidget(self._hypCanvas.canvas.native)
        self._chanGrid.addWidget(self._HypW, len(self) + 2, 1, 1, 1)
        # Add label :
        self._hypLabel = QtWidgets.QWidget()
        layout = QtWidgets.QVBoxLayout(self._hypLabel)
        layout.setContentsMargins(0, 0, 0, 0)
        self._hypYLabels = []
        for k in [''] + self._href + ['']:
            label = QtWidgets.QLabel()
            label.setText(self._addspace + k)
            label.setFont(self._font)
            layout.addWidget(label)
            self._hypYLabels.append(label)
        self._chanGrid.addWidget(self._hypLabel, len(self) + 2, 0, 1, 1)
        PROFILER("Hypnogram", level=2)
        # Connect :
        self._PanHypnoReset.clicked.connect(self._fcn_hypno_clean)
        self._PanHypnoLw.valueChanged.connect(self._fcn_set_hypno_lw)
        self._PanHypnoColor.clicked.connect(self._fcn_set_hypno_color)

        # =====================================================================
        # TOPOPLOT
        # =====================================================================
        # Main canvas for the spectrogram :
        self._topoCanvas = AxisCanvas(axis=False, name='Topoplot')
        self._topoLayout.addWidget(self._topoCanvas.canvas.native)
        self._topoW.setVisible(False)
        self._PanTopoCmin.setValue(-.5)
        self._PanTopoCmax.setValue(.5)
        # Connect :
        self._PanTopoCmap.addItems(self._cmap_lst)
        self._PanTopoCmap.setCurrentIndex(self._cmap_lst.index('Spectral'))
        self._PanTopoCmin.setKeyboardTracking(False)
        self._PanTopoCmin.valueChanged.connect(self._fcn_topo_settings)
        self._PanTopoCmax.setKeyboardTracking(False)
        self._PanTopoRev.clicked.connect(self._fcn_topo_settings)
        self._PanTopoCmax.valueChanged.connect(self._fcn_topo_settings)
        self._PanTopoCmap.currentIndexChanged.connect(self._fcn_topo_settings)
        self._PanTopoDisp.currentIndexChanged.connect(self._fcn_topo_settings)
        self._PanTopoFmin.valueChanged.connect(self._fcn_topo_settings)
        self._PanTopoFmax.valueChanged.connect(self._fcn_topo_settings)
        self._PanTopoAutoClim.clicked.connect(self._fcn_topo_settings)
        self._PanTopoApply.clicked.connect(self._fcn_topo_apply)
        PROFILER("Topoplot", level=2)

        # =====================================================================
        # TIME AXIS
        # =====================================================================
        # Create a unique time axis :
        self._TimeAxis = TimeAxis(axis=self._ax, name='TimeAxis',
                                  indic_color=self._indicol,
                                  fcn=[self.on_mouse_wheel])
        self._TimeAxisW, self._TimeLayout = self._create_compatible_w("TimeW",
                                                                      "TimeL")
        self._TimeLayout.addWidget(self._TimeAxis.canvas.native)
        self._TimeAxisW.setMaximumHeight(400)
        self._TimeAxisW.setMinimumHeight(50)
        self._chanGrid.addWidget(self._TimeAxisW, len(self) + 3, 1, 1, 1)
        # Add label :
        self._timeLabel = QtWidgets.QLabel(self.centralwidget)
        self._timeLabel.setText(self._addspace + 'Time')
        self._timeLabel.setFont(self._font)
        self._chanGrid.addWidget(self._timeLabel, len(self) + 3, 0, 1, 1)
        PROFILER("Time axis", level=2)

    # =====================================================================
    # CHANNELS
    # =====================================================================
    def _fcn_pan_pick(self):
        """Change the panel."""
        idx = int(self._pan_pick.currentIndex())
        self._stacked_panels.setCurrentIndex(idx)

    def _create_compatible_w(self, name_wiget, name_layout, visible=False):
        """Create a widget and a layout."""
        Widget = QtWidgets.QWidget(self.centralwidget)  # noqa
        Widget.setMinimumSize(QtCore.QSize(0, 0))
        Widget.setObjectName(_fromUtf8(name_wiget))
        Widget.setVisible(visible)
        vlay = QtWidgets.QVBoxLayout(Widget)
        vlay.setContentsMargins(9, 0, 9, 0)
        vlay.setSpacing(0)
        vlay.setObjectName(_fromUtf8("vlay"))
        # Create layout :
        Layout = QtWidgets.QVBoxLayout()  # noqa
        Layout.setSpacing(0)
        Layout.setObjectName(_fromUtf8(name_layout))
        vlay.addLayout(Layout)

        return Widget, Layout

    def _fcn_chan_check_and_create_w(self):
        """Create one checkbox and one widget/layout per channel."""
        # Empty list of checkbox and widgets/layouts :
        self._chanChecks = [0] * len(self)
        self._yminSpin, self._ymaxSpin = [0] * len(self), [0] * len(self)
        self._chanWidget = [0] * len(self)
        self._chanLayout = [0] * len(self)
        self._chanCanvas = [0] * len(self)
        self._chanLabels = []
        self._amplitudeTxt = []

        # Define a vertical and horizontal spacers :
        vspacer = QtWidgets.QSpacerItem(20, 40,
                                        QtWidgets.QSizePolicy.Expanding,
                                        QtWidgets.QSizePolicy.Minimum)
        hspacer = QtWidgets.QSpacerItem(40, 20,
                                        QtWidgets.QSizePolicy.Expanding,
                                        QtWidgets.QSizePolicy.Minimum)

        # Loop over channels :
        for i, k in enumerate(self._channels):
            # ============ CHECKBOX ============
            # ----- MAIN CHECKBOX -----
            # Add a checkbox to the scrolling panel :
            self._chanChecks[i] = QtWidgets.QCheckBox(self._PanScrollChan)
            # Name checkbox with channel name :
            self._chanChecks[i].setObjectName(_fromUtf8("_CheckChan" + k))
            self._chanChecks[i].setText(k)
            self._chanChecks[i].setShortcut("Ctrl+" + str(i))
            # Add checkbox to the grid :
            self._PanChanLay.addWidget(self._chanChecks[i], i, 0, 1, 1)
            # Connect with the function :
            self._chanChecks[i].clicked.connect(self._fcn_chan_viz)

            # ----- LABEL/ Y-MIN / Y-MAX -----
            fact = 5.
            # Add amplitude label :
            amplitude = QtWidgets.QLabel(self._PanScrollChan)
            amplitude.setText('Amp')
            self._amplitudeTxt.append(amplitude)
            self._PanChanLay.addWidget(amplitude, i, 2, 1, 1)
            # Add ymin spinbox :
            self._yminSpin[i] = QtWidgets.QDoubleSpinBox(self._PanScrollChan)
            self._yminSpin[i].setDecimals(1)
            self._yminSpin[i].setMinimum(self['min'][i])
            self._yminSpin[i].setMaximum(self['max'][i])
            self._yminSpin[i].setProperty("value", -int(fact * self['std'][i]))
            self._yminSpin[i].setSingleStep(1.)
            self._PanChanLay.addWidget(self._yminSpin[i], i, 3, 1, 1)
            # Add ymax spinbox :
            self._ymaxSpin[i] = QtWidgets.QDoubleSpinBox(self._PanScrollChan)
            self._ymaxSpin[i].setDecimals(1)
            self._ymaxSpin[i].setMinimum(self['min'][i])
            self._ymaxSpin[i].setMaximum(self['max'][i])
            self._ymaxSpin[i].setSingleStep(1.)
            self._ymaxSpin[i].setProperty("value", int(fact * self['std'][i]))
            self._PanChanLay.addWidget(self._ymaxSpin[i], i, 4, 1, 1)
            # Connect buttons :
            self._yminSpin[i].valueChanged.connect(self._fcn_chan_amplitude)
            self._ymaxSpin[i].valueChanged.connect(self._fcn_chan_amplitude)

            # ============ WIDGETS / LAYOUTS ============
            # Create a widget :
            (self._chanWidget[i],
             self._chanLayout[i]) = self._create_compatible_w(
                "_widgetChan" + k, "_LayoutChan" + k)
            self._chanGrid.addWidget(self._chanWidget[i], i, 1, 1, 1)
            # Add channel label :
            self._chanLabels.append(QtWidgets.QLabel(self.centralwidget))
            self._chanLabels[i].setText(self._addspace + k)
            self._chanLabels[i].setFont(self._font)
            self._chanLabels[i].setVisible(False)
            self._chanGrid.addWidget(self._chanLabels[i], i, 0, 1, 1)

            # ============ CANVAS ============
            # Create canvas :
            self._chanCanvas[i] = AxisCanvas(axis=self._ax,
                                             name='Canvas_' + k,
                                             fcn=[self.on_mouse_wheel])
            # Add the canvas to the layout :
            self._chanLayout[i].addWidget(self._chanCanvas[i].canvas.native)

        self._PanChanLay.addItem(vspacer, i + 1, 0, 1, 1)
        self._chanGrid.addItem(hspacer, i + 4, 1, 1, 1)

    # =====================================================================
    # AMPLITUDES
    # =====================================================================
    def _fcn_chan_auto_amp(self):
        """Use automatic amplitudes."""
        viz = not self._PanAmpAuto.isChecked()
        # Set auto-amp :
        self._chan.autoamp = not viz
        # Disable all amplitude related buttons :
        for k in range(len(self._channels)):
            self._amplitudeTxt[k].setEnabled(viz)
            self._yminSpin[k].setEnabled(viz)
            self._ymaxSpin[k].setEnabled(viz)
        self._PanAllAmpMin.setEnabled(viz)
        self._PanAllAmpMax.setEnabled(viz)
        self._PanAmpSym.setEnabled(viz)
        # Finaly, update :
        if viz:
            self._fcn_chan_amplitude()
        else:
            self._chan.update()

    def _fcn_chan_sym_amp(self):
        """Use symetric amplitudes."""
        viz = not self._PanAmpSym.isChecked()
        # Hide amplitude min for all chan :
        for k in range(len(self._channels)):
            self._yminSpin[k].setVisible(viz)
            if not viz:
                self._ymaxSpin[k].setMinimum(.1)
            else:
                self._ymaxSpin[k].setMinimum(self['min'][k])
        self._PanAllAmpMin.setVisible(viz)
        if not viz:
            self._PanAllAmpMax.setMinimum(.1)
        else:
            self._PanAllAmpMax.setMinimum(self['min'].min())
        self._fcn_chan_amplitude()

    def _fcn_chan_amplitude(self):
        """Change amplitude of each channel."""
        # Loop over spinbox and update camera rect :
        for k, (m, M) in enumerate(zip(self._yminSpin, self._ymaxSpin)):
            # Use either symetric / non-symetric amplitudes :
            if self._PanAmpSym.isChecked():
                self._ylims[k, :] = np.array([-M.value(), M.value()])
            else:
                self._ylims[k, :] = np.array([m.value(), M.value()])
            rect = (self._chan.x[0], self._ylims[k, 0],
                    self._chan.x[1] - self._chan.x[0],
                    self._ylims[k, 1] - self._ylims[k, 0])
            self._chanCam[k].rect = rect

    def _fcn_all_amp(self):
        """Set all channel amplitudes."""
        for k, (m, M) in enumerate(zip(self._yminSpin, self._ymaxSpin)):
            m.setValue(self._PanAllAmpMin.value())
            M.setValue(self._PanAllAmpMax.value())
        self._fcn_chan_amplitude()

    def _fcn_update_amp_info(self):
        """Update informations about amplitudes."""
        self._get_data_info()
        self._PanAllAmpMin.setMinimum(self['min'].min())
        self._PanAllAmpMin.setMaximum(self['max'].max())
        self._PanAllAmpMax.setMinimum(self['min'].min())
        self._PanAllAmpMax.setMaximum(self['max'].max())

    def _fcn_chan_set_lw(self):
        """Set channels line-width."""
        self._chan.width = self._channels_lw.value()
        self._fcn_slider_move()

    def _fcn_chan_antialias(self):
        """Set anti-aliasing lines."""
        aa = self._channels_alias.isChecked()
        for k in range(len(self._channels)):
            self._chan.mesh[k].antialias = aa
            self._chan.mesh[k].update()
        if aa:
            self._channels_lw.setMinimum(1.5)
        else:
            self._channels_lw.setMinimum(1.)
        self._fcn_chan_set_lw()

    # =====================================================================
    # CHANNEL SELECTION
    # =====================================================================
    def _fcn_chan_viz(self):
        """Control visible panels of channels."""
        for i, k in enumerate(self._chanChecks):
            viz = k.isChecked()
            self._chanWidget[i].setVisible(viz)
            self._chanLabels[i].setVisible(viz)
            self._chan.visible[i] = viz
            if viz:
                self._chanCanvas[i].set_camera(self._chanCam[i])
        self._chan.update()

    def _fcn_select_all_chan(self):
        """Select all channels."""
        for k in self._chanChecks:
            k.setChecked(True)
        self._fcn_chan_viz()

    def _fcn_deselect_all_chan(self):
        """De-select all channels."""
        for k in self._chanChecks:
            k.setChecked(False)
        self._fcn_chan_viz()

    def _canvas_is_visible(self, k):
        """Find if canvas k is visible.

        Parameters
        ----------
        k : int
            Index of the canvas.

        Returns
        -------
        visible : bool
            A boolean value indicating if the canvas is visible.
        """
        return self._chanWidget[k].isVisible()

    def _canvas_set_visible(self, k, value):
        """Set the visibility of the canvas k to value.

        Parameters
        ----------
        k : int
            Index of the canvas.
        value : bool
            Boolean value if the canvas has to be visible.
        """
        self._chanChecks[k].setChecked(value)
        self._chanWidget[k].setVisible(value)
        self._chanLabels[k].setVisible(value)
        self._chanCanvas[k].set_camera(self._chanCam[k])

    # =====================================================================
    # SPECTROGRAM
    # =====================================================================
    def _fcn_spec_set_data(self):
        """Set data to the spectrogram."""
        # Get nfft and overlap :
        nfft, over = self._PanSpecNfft.value(), self._PanSpecStep.value()
        # Get starting / ending frequency :
        fstart, fend = self._PanSpecFstart.value(), self._PanSpecFend.value()
        # Get contrast :
        contrast = self._PanSpecCon.value()
        contrast = 1. if contrast < .1 else contrast
        # Get colormap :
        cmap = str(self._PanSpecCmap.currentText())
        # Get channel to get spectrogram :
        chan = self._PanSpecChan.currentIndex()
        # Use spectrogram / tf :
        method = str(self._PanSpecMethod.currentText())
        # Interpolation :
        interp = str(self._PanSpecInterp.currentText())
        # Normalization :
        norm = int(self._PanSpecNorm.currentIndex())
        # Get reversed colormap :
        if self._PanSpecCmapInv.isChecked():
            cmap += '_r'
        self._specLabel.setText(self._addspace + self._channels[chan])
        # Set data :
        self._spec.set_data(self._sf, self._data[chan, ...], self._time,
                            nfft=nfft, overlap=over, fstart=fstart, fend=fend,
                            cmap=cmap, contrast=contrast, interp=interp,
                            norm=norm, method=method)
        # Set apply button disable :
        self._PanSpecApply.setEnabled(False)

    def _fcn_spec_compat(self):
        """Check compatibility between spectro parameters."""
        # Get nfft and overlap :
        nfft, _ = self._PanSpecNfft.value(), self._PanSpecStep.value()
        # Get starting / ending frequency :
        _, fend = self._PanSpecFstart.value(), self._PanSpecFend.value()  # noqa
        # Enable / disable normalization :
        use_tf = 1 if str(self._PanSpecMethod.currentText()
                          ) == 'Wavelet' else 0
        self._PanSpecNormW.setEnabled(use_tf)

        self._PanSpecStep.setMaximum(nfft * .99)
        self._PanSpecFend.setMaximum(self._sf / 2)
        self._PanSpecFstart.setMaximum(fend - 0.99)
        # Set apply button enable :
        self._PanSpecApply.setEnabled(True)

    def _fcn_spec_interp(self):
        """Change the 2-D interpolation method."""
        # Interpolation :
        self._spec.interp = str(self._PanSpecInterp.currentText())

    # =====================================================================
    # HYPNOGRAM
    # =====================================================================
    def _fcn_set_hypno_lw(self):
        """Change the line width of the hypnogram."""
        self._hyp.width = self._PanHypnoLw.value()
        self._hyp.set_data(self._sf, self._hypno, self._time)

    def _fcn_set_hypno_color(self):
        """Change the color of the hypnogram."""
        if not(self._PanHypnoColor.isChecked()):
            color = {-1: '#292824', 0: '#292824', 1: '#292824',
                     2: '#292824', 3: '#292824', 4: '#292824'}
        else:
            color = self._hypcolor
        # Get color :
        zp = zip(color.keys(), color.values())
        self._hyp.color = {k: color2vb(color=i) for k, i in zp}
        # Update hypnogram
        self._hyp.set_data(self._sf, self._hypno, self._time)

    def _fcn_hypno_clean(self):
        """Clean the hypnogram."""
        # Confirmation dialog
        reply = QtWidgets.QMessageBox.question(self, 'Message',
                                               "Are you sure you want to reset"
                                               " the program?",
                                               QtWidgets.QMessageBox.Yes,
                                               QtWidgets.QMessageBox.No)

        if reply == QtWidgets.QMessageBox.Yes:
            self._hypno = np.zeros((len(self._hyp),), dtype=np.float32)
            self._hyp.clean(self._sf, self._time)
            # Update info table :
            self._fcn_info_update()
            # Update scoring table :
            self._fcn_hypno_to_score()
            self._fcn_score_to_hypno()
        else:
            pass

    # =====================================================================
    # TOPOPLOT
    # =====================================================================
    def _fcn_topo_settings(self):
        """Manage colormap of the topoplot."""
        # ============== TYPE ==============
        dispas = self._PanTopoDisp.currentText()
        self._topo.filt = True
        self._topo.dispas = dispas
        self._topo.fstart = self._PanTopoFmin.value()
        self._topo.fend = self._PanTopoFmax.value()

        # ============== LIMITS / COLORMAP ==============
        # Get limits :
        if self._PanTopoAutoClim.isChecked():
            clim = None
            self._PanTopoClimW.setEnabled(False)
        else:
            self._PanTopoClimW.setEnabled(True)
            cmin = self._PanTopoCmin.value()
            cmax = self._PanTopoCmax.value()
            clim = (float(cmin), float(cmax))
        # Get and set colormap :
        rv = self._PanTopoRev.isChecked()
        cmap = self._PanTopoCmap.currentText() + rv * '_r'

        # Send data :
        self._topo._cmap = cmap
        self._topo._clim = clim
        self._topo._cblabel = dispas
        self._topo.set_sleep_topo()

        # Finally, enable apply button :
        self._PanTopoApply.setEnabled(True)

    def _fcn_topo_apply(self):
        """Apply topo settings."""
        self._fcn_slider_move()
        self._PanTopoApply.setEnabled(False)



"""Top level Sleep class."""
import logging

import numpy as np

import vispy.scene.cameras as viscam

from .interface import UiInit, UiElements
from .visuals import Visuals
from sleep_scoring._pyqt_module import _PyQtModule
from sleep_scoring.utils import (FixedCam, color2vb, MouseEventControl)
from sleep_scoring.io import ReadSleepData
from sleep_scoring.config import PROFILER

logger = logging.getLogger('visbrain')


class Sleep(_PyQtModule, ReadSleepData, UiInit, Visuals, UiElements,
            MouseEventControl):
    """Visualize and edit sleep data.

    Use this module to :

        * Load and visualize polysomnographic data and spectrogram.
        * Load, edit and save hypnogram from the interface
        * Perform automatic events detection
        * Further signal processing tools (de-mean, de-trend and filtering)
        * Topographic data visualization

    Sleep has been developped in collaboration with Raphael Vallat.

    Parameters
    ----------
    data : string, array_like | None
        Polysomnographic data. Must either be a path to a supported file (see
        notes) or an array of raw data of shape (n_channels, n_pts). If None,
        a dialog window to load the file should appear.
    hypno : array_like | None
        Hypnogram data. Should be a raw vector of shape (n_pts,)
    config_file : string | None
        Path to the configuration file (.txt)
    annotations : string | None
        Path to the annotation file (.txt, .csv). Alternatively, you can pass
        an annotation instance of MNE or simply an (N,) array describing
        the onset.
    channels : list | None
        List of channel names. The length of this list must be n_channels.
    sf : float | None
        The sampling frequency of raw data.
    downsample : float | 100.
        The downsampling frequency for the data and hypnogram raw data.
    axis : bool | False
        Specify if each axis have to contains its own axis. Be carefull
        with this option, the rendering can be much slower.
    href : list | ['art', 'wake', 'rem', 'n1', 'n2', 'n3']
        List of sleep stages. This list can be used to changed the display
        order into the GUI.
    preload : bool | True
        Preload data into memory. For large datasets, turn this parameter to
        True.
    use_mne : bool | False
        Force to load the file using mne.io functions.
    kwargs_mne : dict | {}
        Dictionary to pass to the mne.io loading function.

    Notes
    -----
    .. note::
        * Supported polysomnographic files : by default, Sleep support .vhdr
          (BrainVision), .eeg (Elan), .trc (Micromed) and .edf (European Data
          Format). If mne-python is installed, this default list of supported
          files is extended to .cnt, .egi, .mff, .edf, .bdf, .gdf, .set, .vhdr.
        * Supported hypnogram files : by default, Sleep support .txt, .csv and
          .hyp hypnogram files.

    .. deprecated:: 0.3.4
        Input arguments `file` and `hypno_file` has been deprecated in 0.3.4
        release. Use instead the `data` and `hypno` inputs.
    """

    def __init__(self, data=None, hypno=None, config_file=None,
                 annotations=None, channels=None, sf=None, downsample=100.,
                 axis=True, href=['art', 'wake', 'rem', 'n1', 'n2', 'n3'],
                 preload=True, use_mne=False, kwargs_mne={}, verbose=None):
        """Init."""
        _PyQtModule.__init__(self, verbose=verbose, icon='sleep_icon.svg')
        # ====================== APP CREATION ======================
        UiInit.__init__(self)

        # Set default GUI state :
        self._set_default_state()

        # Mouse control :
        MouseEventControl.__init__(self)

        # ====================== LOAD FILE ======================
        PROFILER("Import file", as_type='title')
        ReadSleepData.__init__(self, data, channels, sf, hypno, href, preload,
                               use_mne, downsample, kwargs_mne,
                               annotations)

        # ====================== VARIABLES ======================
        # Check all data :
        self._config_file = config_file
        self._annot_mark = np.array([])
        self._hconvinv = {v: k for k, v in self._hconv.items()}
        self._ax = axis
        # ---------- Default line width ----------
        self._lw = 1.
        self._lwhyp = 2
        self._defwin = 30.
        self._defstd = 5.
        # ---------- Default colors ----------
        self._chancolor = '#292824'
        # self._hypcolor = '#292824'
        # Hypnogram color :
        self._hypcolor = {-1: '#8bbf56', 0: '#56bf8b', 1: '#aabcce',
                          2: '#405c79', 3: '#0b1c2c', 4: '#bf5656'}
        # Convert color :
        if self._hconv != self._hconvinv:
            hypc = self._hypcolor.copy()
            for k in self._hconv.keys():
                self._hypcolor[k] = hypc[self._hconvinv[k]]
        self._indicol = '#e74c3c'
        # Default spectrogram colormap :
        self._defcmap = 'viridis'
        # Spindles / REM / Peaks colors :
        self._defspin = color2vb('#d73737')
        self._defsw = color2vb('#56bf8b')
        self._defkc = color2vb('#b45a3c')
        self._defrem = color2vb('#6684e1')
        self._defmt = color2vb('#FE8625')
        self._defpeaks = '#b854d4'
        # ---------- Symbol ----------
        self._spinsym = 'x'
        self._swsym = 'o'
        self._kcsym = 'diamond'
        self._remsym = 'triangle_down'
        self._mtsym = 'star'
        self._peaksym = 'disc'
        # ---------- Custom detections ----------
        self._custom_detections = {}
        # Get some data info (min / max / std / mean)
        self._get_data_info()
        PROFILER("Data info")

        # ====================== USER & GUI INTERACTION  ======================
        # User <-> GUI :
        PROFILER("Initialize GUI interactions", as_type='title')
        UiElements.__init__(self)

        # ====================== CAMERAS ======================
        self._cam_creation()

        # ====================== OBJECTS CREATION ======================
        PROFILER("Initialize visual elements", as_type='title')
        Visuals.__init__(self)

        # ====================== FUNCTIONS ON LOAD ======================
        self._fcns_on_creation()
        PROFILER("Functions on creation")

    def __len__(self):
        """Return the number of channels."""
        return len(self._channels)

    def __getitem__(self, key):
        """Return corresponding data info."""
        return self._datainfo[key]

    def replace_detections(self, dtype, method):
        """Replace the default detection methods.

        Parameters
        ----------
        dtype : string
            Name of the method to replace. Should be 'spindle', 'sw' (slow
            waves), 'kc' (k-complexes), 'rem' (rapid eye movements), 'mt'
            (muscle twitches) or 'peak'.
        method : function
            Function to replace the detection. The function should take as an
            input :

                * A vector array of data of shape (n_time_points,)
                * The sampling frequency (float)
                * The time vector of shape (n_time_points,)
                * A vector array for the hypnogram of shape (n_time_points,)

            Then, the function should return indices of relevant events.
            Returned indices should either be :

                *  An array of shape (n_events, 2) where `n_events` describe
                   the number of detected events. The first and second columns
                   of the array respectively describe where detected events
                   start and finished.
                * A boolean vector of shape (n_time_points,) where True values
                  refer to detected events.
                * An array which contains consecutive indices of detected
                  events.

        Examples
        --------
        >>> def method(data, sf, hypno):
        >>>     # Do stuff
        >>>     indices = ...
        >>>     return indices
        """
        # Type checking :
        meth_names = ('spindle', 'sw', 'kc', 'rem', 'mt', 'peak')
        if dtype not in meth_names:
            raise ValueError("dtype should be a string. Choose between "
                             "%s" % ', '.join(meth_names))
        assert hasattr(method, '__call__')
        # Save the method :
        self._custom_detections[dtype] = method
        logger.info("Method for %s detection has been successfully "
                    "replaced" % dtype)
        # Set stack detections enable / disable :
        self._fcn_switch_detection()

    ###########################################################################
    # SUB-FONCTIONS
    ###########################################################################
    def _get_data_info(self):
        """Get some info about data (min, max, std, mean, dist)."""
        self._datainfo = {'min': self._data.min(1), 'max': self._data.max(1),
                          'std': self._data.std(1), 'mean': self._data.mean(1),
                          'dist': self._data.max(1) - self._data.min(1)}

    def _set_default_state(self):
        """Set the default window state."""
        # ================= TAB =================
        self._DetectionTab.setCurrentIndex(0)
        self._stacked_panels.setCurrentIndex(0)
        self._stacked_tools.setCurrentIndex(0)
        self._stacked_detections.setCurrentIndex(0)

    def _cam_creation(self):
        """Create a set of cameras."""
        # ------------------- Channels -------------------
        self._chanCam = []
        for k in range(len(self)):
            self._chanCam.append(FixedCam())  # viscam.PanZoomCamera()
        # ------------------- Spectrogram -------------------
        self._speccam = FixedCam()  # viscam.PanZoomCamera()
        self._specCanvas.set_camera(self._speccam)
        # ------------------- Hypnogram -------------------
        self._hypcam = FixedCam()  # viscam.PanZoomCamera()
        self._hypCanvas.set_camera(self._hypcam)
        # ------------------- Topoplot -------------------
        self._topocam = viscam.PanZoomCamera()
        self._topoCanvas.set_camera(self._topocam)
        # ------------------- Time axis -------------------
        self._timecam = FixedCam()
        self._TimeAxis.set_camera(self._timecam)

        # Keep all cams :
        self._allCams = (self._chanCam, self._speccam, self._hypcam,
                         self._topocam, self._timecam)

    def _fcns_on_creation(self):
        """Applied on creation."""
        self._fcn_grid_toggle()
        self._fcn_slider_move()
        self._chanChecks[0].setChecked(True)
        self._hypLabel.setVisible(self.menuDispHypno.isChecked())
        self._fcn_chan_viz()
        self._fcn_chan_sym_amp()
        self._fcn_info_update()
        self._fcn_hypno_to_score()
        # Set objects visible :
        self._SpecW.setVisible(True)
        self._HypW.setVisible(True)
        self._TimeAxisW.setVisible(True)
        # File to load :
        if self._config_file is not None:  # Config file
            self._load_config(filename=self._config_file)
        if self._annot_file is not None:   # Annotation file
            self._load_annotation_table(filename=self._annot_file)



"""Utility functions for saving/loading files.

- get_file_ext : Get the filename and extension
- safety_save : If a file exist, avoid arasing it when saving.
"""
import os

__all__ = ('get_file_ext', 'safety_save')


def get_file_ext(path):
    """Get the filename and extension.

    Parameters
    ----------
    path : string
        Path to the file.

    Returns
    -------
    file : string
        Name of the file
    ext : string
        Extension of the file.
    """
    # Test if the path exist and if it is a file (not a directory):
    if not os.path.isfile(path):
        # If it is a directory, check if it is a mff directory
        if path[-3:].lower() != "mff":
            raise ValueError("The provided path (" + path + ") is not valid.")
    # Find file extension :
    file, ext = os.path.splitext(path)
    # Be sure to be in lowercase :
    ext = ext.lower()
    return file, ext


def safety_save(path, limit=100):
    """If a file exist, avoid erasing it when saving.

    Parameters
    ----------
    path : string
        Path to the file.
    limit : int | 100
        Limit for the filename occurence.

    Returns
    -------
    name : string
        Unique filename.
    """
    k = 1
    while os.path.isfile(path) and (k < limit):
        fname, fext = os.path.splitext(path)
        if fname.find('(') + 1:
            path = fname[0:fname.find('(') + 1] + str(k) + ')' + fext
        else:
            path = fname + '(' + str(k) + ')' + fext
        k += 1
    return path



"""Visual objects."""
# from .brain_visual import BrainMesh  # noqa
from .cbar import *  # noqa
from .grid_signal_visual import GridSignal  # noqa
from .hypno_visual import Hypnogram  # noqa
from .pic_visual import PicMesh  # noqa
from .tf_map_visual import TFmapsMesh  # noqa
from .topo_visual import TopoMesh  # noqa



"""Import visbrain objects."""
# from .brain_obj import BrainObj  # noqa
from .cbar_obj import ColorbarObj  # noqa
# from .connect_obj import ConnectObj, CombineConnect  # noqa
# from .crossec_obj import CrossSecObj  # noqa
from .image_obj import ImageObj  # noqa
from .hypno_obj import HypnogramObj  # noqa
from .pacmap_obj import PacmapObj  # noqa
# from .picture3d_obj import Picture3DObj, CombinePictures  # noqa
# from .roi_obj import RoiObj, CombineRoi  # noqa
from .scene_obj import SceneObj, VisbrainCanvas  # noqa
# from .source_obj import SourceObj, CombineSources  # noqa
from .tf_obj import TimeFrequencyObj  # noqa
from .topo_obj import TopoObj  # noqa
from .ts3d_obj import TimeSeries3DObj, CombineTimeSeries  # noqa
from .vector_obj import VectorObj, CombineVectors  # noqa
from .visbrain_obj import VisbrainObject, CombineObjects  # noqa
# from .volume_obj import VolumeObj  # noqa



"""Time-frequency maps visual base class.

Authors: Etienne Combrisson <e.combrisson@gmail.com>

License: BSD (3-clause)
"""
import numpy as np

import vispy.visuals.transforms as vist
from vispy.scene.visuals import Image

from ..visuals import CbarBase
from ..utils import (morlet, cmap_to_glsl, averaging, normalization)


__all__ = ('TFmapsMesh')


class TFmapsMesh(CbarBase):
    """Visual class for time-frequency maps.

    Parameters
    ----------
    data : array_like
        Array of data of shape (N,)
    sf : float
        The sampling frequency.
    f_min : float | 1.
        Minimum frequency.
    f_max : float | 160.
        Maximum frequency.
    f_step : float | 2.
        Frequency step between two consecutive frequencies.
    baseline : array_like | None
        Baseline period.
    norm : int | None
        The normalization method.
    """

    def __init__(self, parent=None, interpolation='nearest'):
        """Init."""
        CbarBase.__init__(self)
        self._cblabel = 'Time-frequency map'
        # Visualization of large images can occur GL bugs. So we fix a limit
        # number of time points :
        self._n_limits = 4000
        # Initialize image object :
        pos = np.random.rand(2, 2, 4)
        self._image = Image(parent=parent, interpolation=interpolation)
        self._image.transform = vist.STTransform()
        self._image.set_data(pos)

    def __len__(self):
        """Return the number of time points."""
        return self._n

    def set_data(self, data, sf, f_min=1., f_max=160., f_step=1.,
                 baseline=None, norm=3, contrast=.1, n_window=None,
                 overlap=0., window='flat', **kwargs):
        """Set data to the time frequency map.

        Parameters
        ----------
        data : array_like
            Array of data of shape (N,)
        sf : float
            The sampling frequency.
        f_min : float | 1.
            Minimum frequency.
        f_max : float | 160.
            Maximum frequency.
        f_step : float | 2.
            Frequency step between two consecutive frequencies.
        baseline : array_like | None
            Baseline period.
        norm : int | None
            The normalization method. See the `normalization` function.
        """
        # ======================= CHECKING =======================
        assert isinstance(data, np.ndarray) and data.ndim == 1
        assert isinstance(sf, (int, float))
        assert isinstance(f_min, (int, float))
        assert isinstance(f_max, (int, float))
        assert isinstance(f_step, (int, float))
        # assert isinstance(baseline)

        # ======================= PRE-ALLOCATION =======================
        self._n = len(data)
        freqs = np.arange(f_min, f_max, f_step)  # frequency vector
        time = np.arange(len(self)) / sf
        tf = np.zeros((len(freqs), len(self)), dtype=data.dtype)

        # ======================= COMPUTE TF =======================
        for i, k in enumerate(freqs):
            tf[i, :] = np.square(np.abs(morlet(data, sf, k)))

        # ======================= NORMALIZATION =======================
        normalization(tf, norm=norm, baseline=baseline, axis=1)

        # ======================= AVERAGING =======================
        if isinstance(n_window, int):
            tf = averaging(tf, n_window, axis=1, overlap=overlap,
                           window=window)

        # ======================= DOWNSAMPLE =======================
        # Downsample large images :
        if tf.shape[1] > self._n_limits:
            downsample = int(np.round(tf.shape[1] / self._n_limits))
            tf = tf[:, ::downsample]

        # ======================= CLIM // CMAP =======================
        # Get contrast (if defined) :
        self._clim = kwargs.get('clim', None)
        if isinstance(contrast, (int, float)) and (self._clim is None):
            self._clim = (tf.min() * contrast, tf.max() * contrast)
        if self._clim is None:
            self._clim = (tf.min(), tf.max())
        kwargs['clim'] = self._clim
        # Cmap :
        self._cmap = kwargs.get('cmap', 'viridis')

        # ======================= COLOR =======================
        self._image.set_data(tf)
        self._image.clim = 'auto'
        self._image.cmap = cmap_to_glsl(limits=(tf.min(), tf.max()), **kwargs)

        # ======================= SCALE // TRANSLATE =======================
        # Scale and translate TF :
        t_min, t_max = time.min(), time.max()
        fr_min, fr_max = freqs.min(), freqs.max()
        # Re-scale the mesh for fitting in time / frequency :
        fact = (fr_max - fr_min) / len(freqs)
        sc = (t_max / tf.shape[1], fact, 1)
        tr = [0., fr_min, 0.]
        self._image.transform.scale = sc
        self._image.transform.translate = tr

        # ======================= CAMERA =======================
        self.rect = (time[0], f_min, t_max - t_min, fr_max - fr_min)
        self.freqs = freqs

    def update(self):
        """Update image."""
        self._image.update()

    # ----------- VISIBLE -----------
    @property
    def visible(self):
        """Get the visible value."""
        return self._visible

    @visible.setter
    def visible(self, value):
        """Set visible value."""
        self._visible = value
        self._image.visible = value

    # ----------- INTERPOLATION -----------
    @property
    def interpolation(self):
        """Get the interpolation value."""
        return self._interpolation

    @interpolation.setter
    def interpolation(self, value):
        """Set interpolation value."""
        self._interpolation = value
        self._image.interpolation = value



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/etienne/Toolbox/visbrain/visbrain/utils/gui/screenshot_gui.ui'
#
# Created by: PyQt5 UI code generator 5.6
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Screenshot(object):
    def setupUi(self, Screenshot):
        Screenshot.setObjectName("Screenshot")
        Screenshot.resize(341, 563)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(Screenshot)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.gridLayout_5 = QtWidgets.QGridLayout()
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.label = QtWidgets.QLabel(Screenshot)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.gridLayout_5.addWidget(self.label, 0, 0, 1, 1)
        self._ssSelect = QtWidgets.QComboBox(Screenshot)
        self._ssSelect.setObjectName("_ssSelect")
        self._ssSelect.addItem("")
        self._ssSelect.addItem("")
        self.gridLayout_5.addWidget(self._ssSelect, 0, 2, 1, 1)
        self.line = QtWidgets.QFrame(Screenshot)
        self.line.setFrameShape(QtWidgets.QFrame.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.gridLayout_5.addWidget(self.line, 0, 1, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_5.addItem(spacerItem, 1, 2, 1, 1)
        self.verticalLayout_3.addLayout(self.gridLayout_5)
        self._ssCanvasW = QtWidgets.QWidget(Screenshot)
        self._ssCanvasW.setObjectName("_ssCanvasW")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self._ssCanvasW)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setObjectName("gridLayout_2")
        self._ssCustomW = QtWidgets.QWidget(self._ssCanvasW)
        self._ssCustomW.setObjectName("_ssCustomW")
        self.verticalLayout = QtWidgets.QVBoxLayout(self._ssCustomW)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout_3 = QtWidgets.QGridLayout()
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.label_5 = QtWidgets.QLabel(self._ssCustomW)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_5.setFont(font)
        self.label_5.setObjectName("label_5")
        self.gridLayout_3.addWidget(self.label_5, 0, 0, 1, 1)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self._ssWidth = QtWidgets.QDoubleSpinBox(self._ssCustomW)
        self._ssWidth.setMaximum(10000.0)
        self._ssWidth.setProperty("value", 10.0)
        self._ssWidth.setObjectName("_ssWidth")
        self.gridLayout.addWidget(self._ssWidth, 0, 1, 1, 1)
        self.label_6 = QtWidgets.QLabel(self._ssCustomW)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 0, 0, 1, 1)
        self.label_7 = QtWidgets.QLabel(self._ssCustomW)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 1, 0, 1, 1)
        self._ssHeight = QtWidgets.QDoubleSpinBox(self._ssCustomW)
        self._ssHeight.setMaximum(10000.0)
        self._ssHeight.setProperty("value", 10.0)
        self._ssHeight.setObjectName("_ssHeight")
        self.gridLayout.addWidget(self._ssHeight, 1, 1, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(40, 0, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem1, 2, 1, 1, 1)
        self.gridLayout_3.addLayout(self.gridLayout, 0, 2, 1, 1)
        self.label_3 = QtWidgets.QLabel(self._ssCustomW)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.gridLayout_3.addWidget(self.label_3, 2, 0, 1, 1)
        self.line_6 = QtWidgets.QFrame(self._ssCustomW)
        self.line_6.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_6.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_6.setObjectName("line_6")
        self.gridLayout_3.addWidget(self.line_6, 2, 1, 1, 1)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_3.addItem(spacerItem2, 5, 2, 1, 1)
        self._ssBgcolor = QtWidgets.QLineEdit(self._ssCustomW)
        self._ssBgcolor.setEnabled(False)
        self._ssBgcolor.setObjectName("_ssBgcolor")
        self.gridLayout_3.addWidget(self._ssBgcolor, 3, 2, 1, 1)
        self._ssUnit = QtWidgets.QComboBox(self._ssCustomW)
        self._ssUnit.setObjectName("_ssUnit")
        self._ssUnit.addItem("")
        self._ssUnit.addItem("")
        self._ssUnit.addItem("")
        self._ssUnit.addItem("")
        self.gridLayout_3.addWidget(self._ssUnit, 1, 2, 1, 1)
        self._ssDpi = QtWidgets.QSpinBox(self._ssCustomW)
        self._ssDpi.setMinimum(100)
        self._ssDpi.setMaximum(600)
        self._ssDpi.setProperty("value", 300)
        self._ssDpi.setObjectName("_ssDpi")
        self.gridLayout_3.addWidget(self._ssDpi, 2, 2, 1, 1)
        self.line_5 = QtWidgets.QFrame(self._ssCustomW)
        self.line_5.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_5.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_5.setObjectName("line_5")
        self.gridLayout_3.addWidget(self.line_5, 1, 1, 1, 1)
        self.line_4 = QtWidgets.QFrame(self._ssCustomW)
        self.line_4.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_4.setObjectName("line_4")
        self.gridLayout_3.addWidget(self.line_4, 0, 1, 1, 1)
        self.line_8 = QtWidgets.QFrame(self._ssCustomW)
        self.line_8.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_8.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_8.setObjectName("line_8")
        self.gridLayout_3.addWidget(self.line_8, 3, 1, 1, 1)
        self.label_4 = QtWidgets.QLabel(self._ssCustomW)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.gridLayout_3.addWidget(self.label_4, 1, 0, 1, 1)
        self._ssTransp = QtWidgets.QCheckBox(self._ssCustomW)
        self._ssTransp.setObjectName("_ssTransp")
        self.gridLayout_3.addWidget(self._ssTransp, 4, 0, 1, 3)
        self._ssBgcolorChk = QtWidgets.QCheckBox(self._ssCustomW)
        font = QtGui.QFont()
        font.setItalic(True)
        self._ssBgcolorChk.setFont(font)
        self._ssBgcolorChk.setObjectName("_ssBgcolorChk")
        self.gridLayout_3.addWidget(self._ssBgcolorChk, 3, 0, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout_3)
        self.gridLayout_2.addWidget(self._ssCustomW, 3, 0, 1, 3)
        self._ssFactorW = QtWidgets.QWidget(self._ssCanvasW)
        self._ssFactorW.setObjectName("_ssFactorW")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self._ssFactorW)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.gridLayout_4 = QtWidgets.QGridLayout()
        self.gridLayout_4.setObjectName("gridLayout_4")
        spacerItem3 = QtWidgets.QSpacerItem(40, 0, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_4.addItem(spacerItem3, 1, 2, 1, 1)
        self.label_9 = QtWidgets.QLabel(self._ssFactorW)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_9.setFont(font)
        self.label_9.setObjectName("label_9")
        self.gridLayout_4.addWidget(self.label_9, 0, 0, 1, 1)
        self._ssFactor = QtWidgets.QDoubleSpinBox(self._ssFactorW)
        self._ssFactor.setMinimum(0.1)
        self._ssFactor.setMaximum(20.0)
        self._ssFactor.setProperty("value", 1.0)
        self._ssFactor.setObjectName("_ssFactor")
        self.gridLayout_4.addWidget(self._ssFactor, 0, 2, 1, 1)
        self.line_7 = QtWidgets.QFrame(self._ssFactorW)
        self.line_7.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_7.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_7.setObjectName("line_7")
        self.gridLayout_4.addWidget(self.line_7, 0, 1, 1, 1)
        self.verticalLayout_4.addLayout(self.gridLayout_4)
        self.gridLayout_2.addWidget(self._ssFactorW, 4, 0, 1, 3)
        self._ssCanvasName = QtWidgets.QComboBox(self._ssCanvasW)
        self._ssCanvasName.setObjectName("_ssCanvasName")
        self.gridLayout_2.addWidget(self._ssCanvasName, 0, 2, 1, 1)
        self._ssResolution = QtWidgets.QComboBox(self._ssCanvasW)
        self._ssResolution.setObjectName("_ssResolution")
        self._ssResolution.addItem("")
        self._ssResolution.addItem("")
        self.gridLayout_2.addWidget(self._ssResolution, 2, 2, 1, 1)
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_2.addItem(spacerItem4, 1, 2, 1, 1)
        self._ssAutocrop = QtWidgets.QCheckBox(self._ssCanvasW)
        self._ssAutocrop.setObjectName("_ssAutocrop")
        self.gridLayout_2.addWidget(self._ssAutocrop, 5, 0, 1, 3)
        self.label_8 = QtWidgets.QLabel(self._ssCanvasW)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_8.setFont(font)
        self.label_8.setObjectName("label_8")
        self.gridLayout_2.addWidget(self.label_8, 2, 0, 1, 1)
        self.label_2 = QtWidgets.QLabel(self._ssCanvasW)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.gridLayout_2.addWidget(self.label_2, 0, 0, 1, 1)
        self.line_2 = QtWidgets.QFrame(self._ssCanvasW)
        self.line_2.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.gridLayout_2.addWidget(self.line_2, 0, 1, 1, 1)
        self.line_3 = QtWidgets.QFrame(self._ssCanvasW)
        self.line_3.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.gridLayout_2.addWidget(self.line_3, 2, 1, 1, 1)
        self.verticalLayout_2.addLayout(self.gridLayout_2)
        self.verticalLayout_3.addWidget(self._ssCanvasW)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem5 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem5)
        self._ssRun = QtWidgets.QPushButton(Screenshot)
        self._ssRun.setObjectName("_ssRun")
        self.horizontalLayout_2.addWidget(self._ssRun)
        spacerItem6 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem6)
        self.verticalLayout_3.addLayout(self.horizontalLayout_2)
        spacerItem7 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem7)

        self.retranslateUi(Screenshot)
        QtCore.QMetaObject.connectSlotsByName(Screenshot)

    def retranslateUi(self, Screenshot):
        _translate = QtCore.QCoreApplication.translate
        Screenshot.setWindowTitle(_translate("Screenshot", "Form"))
        self.label.setText(_translate("Screenshot", "Render"))
        self._ssSelect.setToolTip(_translate("Screenshot", "<html><head/><body><p>Render either :</p><p>- The entire window</p><p>- One selected canvas</p></body></html>"))
        self._ssSelect.setItemText(0, _translate("Screenshot", "Selected canvas"))
        self._ssSelect.setItemText(1, _translate("Screenshot", "Entire window"))
        self.label_5.setText(_translate("Screenshot", "Size"))
        self._ssWidth.setToolTip(_translate("Screenshot", "<html><head/><body><p>Width (in centimeter, millimeter, pixel or inch) of rendered image at the dpi level</p></body></html>"))
        self.label_6.setText(_translate("Screenshot", "Width"))
        self.label_7.setText(_translate("Screenshot", "Height"))
        self._ssHeight.setToolTip(_translate("Screenshot", "<html><head/><body><p>Height (in centimeter, millimeter, pixel or inch) of rendered image at the dpi level</p></body></html>"))
        self.label_3.setText(_translate("Screenshot", "dpi"))
        self._ssBgcolor.setToolTip(_translate("Screenshot", "<html><head/><body><p>Use either the name of the color (e.g. black, white...), a RGB tuple (e.g. (0., 1., 0.)...) or an hexadecimal string (e.g. #ab4642)</p></body></html>"))
        self._ssBgcolor.setPlaceholderText(_translate("Screenshot", "black, (.1, .1, .1), #ab4642..."))
        self._ssUnit.setToolTip(_translate("Screenshot", "<html><head/><body><p>Size unit</p></body></html>"))
        self._ssUnit.setItemText(0, _translate("Screenshot", "centimeter"))
        self._ssUnit.setItemText(1, _translate("Screenshot", "millimeter"))
        self._ssUnit.setItemText(2, _translate("Screenshot", "pixel"))
        self._ssUnit.setItemText(3, _translate("Screenshot", "inch"))
        self._ssDpi.setToolTip(_translate("Screenshot", "<html><head/><body><p>dpi level for the rendered image</p></body></html>"))
        self.label_4.setText(_translate("Screenshot", "Unit"))
        self._ssTransp.setToolTip(_translate("Screenshot", "<html><head/><body><p>Force to have a transparent background</p></body></html>"))
        self._ssTransp.setText(_translate("Screenshot", "Use transparent background"))
        self._ssBgcolorChk.setToolTip(_translate("Screenshot", "<html><head/><body><p><span style=\" font-style:normal;\">Change the background color.</span></p></body></html>"))
        self._ssBgcolorChk.setText(_translate("Screenshot", "Background\n"
"color"))
        self.label_9.setText(_translate("Screenshot", "Factor"))
        self._ssFactor.setToolTip(_translate("Screenshot", "<html><head/><body><p>Multiply the original canvas size by this factor. For instance, if the canvas has a size of (1020, 800) and factor is 2. then the rendered image will have a size of (2040, 1600)</p></body></html>"))
        self._ssCanvasName.setToolTip(_translate("Screenshot", "<html><head/><body><p>Choose the canvas to render</p></body></html>"))
        self._ssResolution.setToolTip(_translate("Screenshot", "<html><head/><body><p>Change the image resolution for rendering :</p><p>- <span style=\" font-weight:600;\">Custom  :</span> specify the image size (controlled by unit) at a specific dpi level. Note that the rendering might not have the exact desired size but will try instead to conserve the proportion width/height</p><p>- <span style=\" font-weight:600;\">Proportional :</span> increase the size of the image proportionaly to the original canvas size.</p></body></html>"))
        self._ssResolution.setItemText(0, _translate("Screenshot", "Custom image size"))
        self._ssResolution.setItemText(1, _translate("Screenshot", "Proportional to screen size"))
        self._ssAutocrop.setToolTip(_translate("Screenshot", "<html><head/><body><p>Automatically crop the image to non-background values</p></body></html>"))
        self._ssAutocrop.setText(_translate("Screenshot", "autocrop"))
        self.label_8.setText(_translate("Screenshot", "Image resolution"))
        self.label_2.setText(_translate("Screenshot", "Canvas name"))
        self._ssRun.setToolTip(_translate("Screenshot", "<html><head/><body><p>Run the screenshot</p></body></html>"))
        self._ssRun.setText(_translate("Screenshot", "Run"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Screenshot = QtWidgets.QWidget()
    ui = Ui_Screenshot()
    ui.setupUi(Screenshot)
    Screenshot.show()
    sys.exit(app.exec_())

